# -*- coding: utf-8 -*-
# Copyright (C) 1999-2015 Mag. Christian Tanzer. All rights reserved
# Glasauergasse 32, A--1130 Wien, Austria. tanzer@swing.co.at
# ****************************************************************************
# This module is part of the package ATAX.
#
# This module is licensed under the terms of the BSD 3-Clause License
# <http://www.c-tanzer.at/license/bsd_3c.html>.
# ****************************************************************************
#
#++
# Name
#    kassabuch
#
# Purpose
#    Write kassabuch file for a month
#
# Revision Dates
#     9-Feb-1999 (CT) Creation
#     6-Feb-2000 (CT) `vst_korrektur' passed to `V_Account' instead of to
#                     `umsatzsteuer'
#     5-Jul-2006 (CT) Import from `_ATAX`
#    17-Sep-2007 (CT) Modernized
#    17-Sep-2007 (CT) Use `Account.add_file`
#    17-Sep-2007 (CT) `main` refactored
#     3-Jan-2010 (CT) Use `TFL.CAO` instead of `TFL.Command_Line`
#    29-Oct-2015 (CT) Improve Python 3 compatibility
#    ««revision-date»»···
#--

from   __future__  import absolute_import
from   __future__  import division
from   __future__  import print_function
from   __future__  import unicode_literals

from   _TFL.pyk          import pyk

from _ATAX.accounting    import *

par_sub_pat = re.compile ("\(([^)]+)\)")

def kassabuch (account, file) :
    """Write entries in `account' to kassabuch `file'"""
    for e in account.entries :
        if not isinstance (e, Account_Entry) :
            continue
        e.desc = par_sub_pat.sub ("[\\1]", e.desc)
        if "-" in e.dir :
            file.write \
                ( pyk.encoded
                    ( "\\A %6s %-50s (%8s) [%s]\n"
                    % (e.date, e.desc, e.gross.as_string (), e.number)
                    )
                )
        elif "+" in e.dir :
            print ("Income entries not yet implemented", e)
# end def kassabuch

class Command (Command) :

    def _create_account (self, cmd, categories, source_currency, vst_korrektur) :
        return V_Account (vst_korrektur = vst_korrektur)
    # end def _create_account

    def _output (self, cmd, account, categories, source_currency) :
        if cmd.output :
            file = open (cmd.output, "wb")
        else :
            file = sys.stdout
        kassabuch (account, file)
    # end def _output

    def _opt_spec (self) :
        return self.__super._opt_spec () + ("-output:S", )
    # end def _opt_spec

# end class Command

if __name__ == "__main__":
    Command ()
### __END__ ATAX.kassabuch



# -*- coding: utf-8 -*-
# Copyright (C) 2002-2015 Mag. Christian Tanzer. All rights reserved
# Glasauergasse 32, A--1130 Wien, Austria. tanzer@swing.co.at
# ****************************************************************************
# This module is part of the package ATAX.
#
# This module is licensed under the terms of the BSD 3-Clause License
# <http://www.c-tanzer.at/license/bsd_3c.html>.
# ****************************************************************************
#
#++
# Name
#    anlagenverzeichnis
#
# Purpose
#    Manage depreciations (as required by Austrian tax law)
#
# Revision Dates
#    12-Feb-2002 (CT) Creation (rewrite of perl code)
#    19-Jan-2003 (CT) Rest value made dependent on target_currency, instead
#                     of source_currency
#    19-Jan-2003 (CT) `-Start_year` added
#    19-Jan-2003 (CT) Don't show entries totally depreciated
#    22-Dec-2003 (CT) `Anlagen_Entry.evaluate` changed to not add
#                     non-contemporary entries to totals
#    19-Feb-2006 (CT) Import from packages _ATAX and _TFL
#    11-Feb-2007 (CT) `string` functions replaced by `str` methods
#    17-Sep-2007 (CT) Use `EUC_Opt_SC` and `EUC_Opt_TC` instead of home-grown
#                     code
#     5-Feb-2008 (CT) Support for `FBiG` added
#    18-Feb-2008 (CT) s/_Base/_Base_/g
#    18-Feb-2008 (CT) `_IFB_`, `IFB`, and `FBiG` factored
#    12-Aug-2008 (CT) `p_konto` added
#    13-Aug-2008 (CT) `_write_entry` changed to only write `e.ifb.value` if
#                     `ifb` is still alive (and output formatting corrected)
#    11-Mar-2009 (CT)  Check `target_currency.to_euro_factor == 1.0` instead
#                      of comparing `target_currency` to `EU_Currency`
#    27-Mar-2009 (CT) `active` factored (and changed to allow entries with
#                     `base_rate == 0`)
#    27-Mar-2009 (CT) `account_format` and `_update_account_entry` changed to
#                     allow line-specific date (for `Abgang`)
#     3-Jan-2010 (CT) Use `TFL.CAO` instead of `TFL.Command_Line`
#     7-Feb-2011 (CT) `cat` and `total_per_cat` added
#     7-Jun-2012 (CT) Use `TFL.r_eval`
#     3-Jan-2014 (CT) Add `output_encoding`; use `pyk.fprint`, not `print`
#    29-Oct-2015 (CT) Improve Python 3 compatibility
#    ««revision-date»»···
#--

from   __future__       import division, print_function
from   __future__       import absolute_import, unicode_literals

from   _ATAX.accounting import *
from   _ATAX.accounting import _Base_, _Entry_

from   _TFL.pyk         import pyk
from   _TFL.Regexp      import *

import _TFL.CAO
import _TFL.r_eval

class _Mixin_ (TFL.Meta.Object) :

    def _setup_dates (self, target_year) :
        self.head_date       = "1.1.%s"   % (target_year, )
        self.midd_date       = "30.6.%s"  % (target_year, )
        self.tail_date       = "31.12.%s" % (target_year, )
        self.head_time       = Date (self.head_date)
        self.midd_time       = Date (self.midd_date)
        self.tail_time       = Date (self.tail_date)
        self.target_year     = int  (target_year)
    # end def _setup_dates

# end class _Mixin_

@pyk.adapt__bool__
class _IFB_ (TFL.Meta.Object) :
    """Base class for FBiG and IFB."""

    def __init__ (self, entry) :
        self.entry  = entry
        self.alive  = entry.birth_time.year + 4 >  entry.target_year
        self.is_new = entry.birth_time.year     == entry.target_year
    # end def __init__

    def round (self) :
        self.value  = self.value.rounded_as_target ()
    # end def round

    def __bool__ (self) :
        return self.alive and bool (self.value)
    # end def __bool__

# end class _IFB_

class FBiG (_IFB_) :
    """Model a FBiG (Freibetrag für investierte Gewinne)."""

    abbr    = "FBiG"
    account = None
    name    = "Freibetrag für investierte Gewinne"

    def __init__ (self, entry, ifb, source_currency) :
        self.__super.__init__ (entry)
        self.value = source_currency (float (ifb or entry.birth_value))
    # end def __init__

# end class FBiG

class IFB (_IFB_) :
    """Model a IFB (Investitionsfreibetrag)."""

    abbr    = "IFB"
    account = 7802
    name    = "Investitionsfreibetrag"

    def __init__ (self, entry, ifb, source_currency) :
        self.__super.__init__ (entry)
        self.rate  = int (ifb or 0) / 100.0
        self.value = entry.birth_value * self.rate
    # end def __init__

# end class IFB

class Anlagen_Entry (_Mixin_, _Entry_) :

    cat            = "Normal"
    rate_pattern   = r"(?P<rate> [-+*/().0-9\s]+)"
    first_rate_pat = Regexp (rate_pattern, re.X)
    later_rate_pat = Regexp \
        ( r"(?P<year> \d\d (?: \d\d)?) \s* : \s* " + rate_pattern
        , re.X
        )
    _cat_pat       = Regexp (r"C\[(?P<cat> [^]]+)\]", re.VERBOSE)

    def __init__ (self, line, anlagenverzeichnis) :
        try :
            ( self.desc, self.supplier, self.flags
            , self.birth_date, self.a_value, self.afa_spec, ifb
            , self.death_date
            )                = split_pat.split     (line, 8)
        except ValueError as exc :
            print (line)
            raise
        final                = "31.12.2037"
        self.p_konto         = self._get_p_konto (self.flags)
        self.birth_time      = Date (self.birth_date)
        self.death_time      = Date (self.death_date or final)
        self.alive           = self.death_time > anlagenverzeichnis.tail_time
        self.contemporary    = \
            (   self.birth_time <= anlagenverzeichnis.tail_time
            and self.death_time >= anlagenverzeichnis.head_time
            )
        if int (self.death_time.year) < int (anlagenverzeichnis.year) :
            self._setup_dates (self.death_time.year)
        else :
            self._setup_dates (anlagenverzeichnis.year)
        self.half_date       = "1.7.%s" % (self.birth_time.year, )
        if "~" in self.flags :
            self.half_date   = "1.1.%s" % (self.birth_time.year + 1, )
        self.half_time       = Date (self.half_date)
        self.desc            = desc_strip_pat.sub  ("", self.desc)
        currency_match       = currency_pat.search (self.a_value)
        a_value              = self.a_value
        source_currency      = anlagenverzeichnis.source_currency
        if currency_match :
            source_currency  = EU_Currency.Table   [currency_match.group (1)]
            a_value          = currency_pat.sub    ("", a_value)
        if EUC.target_currency is not ATS :
            self.zero        = source_currency     (0.0)
        else :
            self.zero        = source_currency     (1.0)
        self.source_currency = source_currency
        self.birth_value     = source_currency     (TFL.r_eval (a_value))
        self.new_value       = source_currency     (0.0)
        self.out_value       = source_currency     (0.0)
        if "G" in self.flags :
            self.ifb         = FBiG (self, ifb, source_currency)
        else :
            self.ifb         = IFB  (self, ifb, source_currency)
        self._set_cat (self.flags)
    # end def __init__

    @property
    def active (self) :
        return \
            (   self.contemporary
            and (self.current_depreciation > 0 or self.base_rate == 0)
            )
    # end def active

    def evaluate (self) :
        self._calc_rates ()
        self.current_depreciation = \
            self.birth_value * (self.current_rate / 100.0)
        if "=" not in self.flags :
            self.head_value = max \
                ( self.birth_value * ((100.0 - self.past_total_rate) / 100.)
                , self.zero
                )
            self.tail_value = self.head_value - self.current_depreciation
            if self.tail_value < self.zero :
                self.tail_value = self.zero
                self.current_depreciation -= self.zero
        else :
            self.head_value = self.tail_value = self.birth_value
        if self.birth_time >= self.head_time :
            self.head_value = self.source_currency (0.0)
            self.new_value  = self.birth_value
        if not self.alive :
            self.out_value  = self.tail_value
            self.tail_value = self.source_currency (0.0)
        if self.tail_value.target_currency.to_euro_factor != 1.0 :
            self.birth_value          = self.birth_value.rounded_as_target ()
            self.head_value           = self.head_value.rounded_as_target  ()
            self.tail_value           = self.tail_value.rounded_as_target  ()
            self.new_value            = self.new_value.rounded_as_target   ()
            self.out_value            = self.out_value.rounded_as_target   ()
            self.current_depreciation = \
                self.current_depreciation.rounded_as_target ()
            if self.ifb :
                self.ifb.round ()
    # end def evaluate

    def _calc_rates (self) :
        rates          = [x.strip () for x in self.afa_spec.split (",")]
        first_rate     = rates [0]
        first_rate_pat = self.first_rate_pat
        later_rate_pat = self.later_rate_pat
        if not first_rate_pat.match (first_rate) :
            raise ValueError \
                ("%s doesn't match a depreciation rate" % (first_rate, ))
        later_rates = []
        for r in rates [1:] :
            if not later_rate_pat.match (r) :
                raise ValueError \
                    ("%s doesn't match a depreciation rate" % (r, ))
            y = Time_Tuple (later_rate_pat.year).year
            later_rates.append ((y, TFL.r_eval (later_rate_pat.rate) * 1.0))
        y_rate  = self.base_rate = TFL.r_eval (first_rate_pat.rate) * 1.0
        if later_rates :
            later_rates.append ((self.target_year, later_rates [-1] [1]))
        else :
            later_rates.append ((self.target_year, y_rate))
        y_rates = self.y_rates   = \
            [y_rate * ((0.5, 1.0) [self.birth_time < self.half_time])]
        if self.birth_time < self.head_time :
            current_year = self.birth_time.year + 1
            for target_year, next_rate in later_rates :
                while current_year < target_year :
                    y_rates.append (y_rate)
                    current_year += 1
                y_rate = self.base_rate = next_rate
            y_rates.append \
                (y_rate * ((0.5, 1.0) [self.midd_time < self.death_time]))
        self.current_rate     = y_rates [-1]
        past_total_rate       = 0
        for y_rate in y_rates [:-1] :
            past_total_rate  += y_rate
        self.past_total_rate  = min (past_total_rate, 100.0)
        if self.past_total_rate + self.current_rate > 100.0 :
            self.current_rate = 100.0 - self.past_total_rate
    # end def _calc_rates

    def _set_cat (self, flags) :
        pat = self._cat_pat
        if pat.search (flags) :
            self.cat = pat.cat
    # end def _set_cat

# end class Anlagen_Entry

class Anlagenverzeichnis (_Mixin_, _Base_) :

    assignment_pat = Regexp \
        ( r"^\$ "
          r"(?P<var> account_file | source_currency)"
          r"\s* = \s*"
          r"(?P<value> .*)"
          r"\s* ;"
        , re.X
        )

    header_format  = "%-48s  %-8s  %10s  %10s  %8s  %10s  %10s"
    entry1_format  = "%-44s%4s  %8s  %10.2f  %10.2f     %5.2f  %10.2f  %10.2f"
    newifb_format  = "  %-46s  %8s  %10s  %10s  %8s  %10.2f  %10s"
    alive_format   = "  %-46s  %8s  %10s  %10s  %8s"
    dying_format   = "  %-36.31s%10s  %8s  %10s  %10s  %8s  %10.2f  %10s"
    footer_format  = "\n%-48s  %8s  %10.2f  %10.2f  %8s  %10.2f  %10.2f"
    new_format     = "%-48s  %8s  %10s  %10.2f"
    out_format     = "%-48s  %8s  %10s  %10s  %8s  %10.2f"

    account_format = \
        " %s & & & %10.2f & b & %-5s & 2100 & - & %-3s & & %-6s %s\n"


    ifb_type       = ""

    def __init__ (self, year, start_year, file_name, source_currency, account_file = None) :
        self.year               = year
        self.start_time         = Date ("1.1.%s" % (start_year, ))
        self.file_name          = file_name
        self.source_currency    = source_currency
        self.account_file       = account_file
        self.entries            = []
        self.total_birth_value  = source_currency (0.0)
        self.total_head_value   = source_currency (0.0)
        self.total_tail_value   = source_currency (0.0)
        self.total_new_value    = source_currency (0.0)
        self.total_out_value    = source_currency (0.0)
        self.total_ifb_value    = source_currency (0.0)
        self.total_depreciation = source_currency (0.0)
        self.total_per_cat      = defaultdict     (EUR)
        self._setup_dates (year)
        self.add_file     (file_name)
    # end def __init__

    def add_file (self, file_name) :
        assignment_pat = self.assignment_pat
        file           = open (file_name, "rb")
        for line in file :
            line                      = self._decoded (line)
            if ignor_pat.match (line) : continue
            line                      = ws_head_pat.sub ("", line, count = 1)
            line                      = ws_tail_pat.sub ("", line, count = 1)
            if not line               : continue
            if assignment_pat.match (line) :
                self.eval_line (line, assignment_pat)
            else :
                self.add_line  (line)
        file.close ()
    # end def add_file

    def eval_line (self, line, match) :
        name       = match.var
        expression = match.value.replace \
                         ("$target_year", str (self.target_year))
        value      = TFL.r_eval (expression)
        if name == "source_currency" :
            value  = EUC.Table [value]
        setattr (self, name, value)
    # end def eval_line

    def add_line (self, line) :
        self.entries.append (Anlagen_Entry (line, self))
    # end def add_line

    def evaluate (self) :
        for e in self.entries :
            if (not e.contemporary) or e.birth_time < self.start_time :
                e.contemporary = 0
                continue
            e.evaluate ()
            if e.active :
                self.total_birth_value     += e.birth_value
                self.total_head_value      += e.head_value
                self.total_tail_value      += e.tail_value
                self.total_new_value       += e.new_value
                self.total_out_value       += e.out_value
                self.total_depreciation    += e.current_depreciation
                self.total_per_cat [e.cat] += e.current_depreciation
            if e.ifb and e.ifb.is_new :
                self.ifb_type               = e.ifb.__class__
                self.total_ifb_value       += e.ifb.value
    # end def evaluate

    def write (self) :
        pyk.fprint \
            ( self.header_format
            % ( "", "", "Anschaff/", "Buchwert", " Afa ", "Afa", "Buchwert")
            )
        pyk.fprint \
            ( self.header_format
            % ( "Text", "Datum", "Teil-Wert", "1.1.", "  %  "
              , "IFB/Abgang", "31.12."
              )
            )
        pyk.fprint ("\n%s\n" % ("=" * 116, ))
        for e in self.entries :
            if e.active :
                self._write_entry (e)
        pyk.fprint ("\n%s\n" % ("=" * 116, ))
        pyk.fprint \
            ( self.footer_format
            % ( "Summe", ""
              , self.total_birth_value
              , self.total_head_value
              , "Afa"
              , self.total_depreciation
              , self.total_tail_value
              )
            )
        if len (self.total_per_cat) > 1 :
            for k, v in sorted (pyk.iteritems (self.total_per_cat)) :
                pyk.fprint ((self.out_format % ("", "", "", "", k, v)))
        pyk.fprint \
            (self.new_format % ("Neuzugänge", "", "", self.total_new_value))
        pyk.fprint \
            ( self.out_format
            % ("Abgänge", "", "", "", "", self.total_out_value)
            )
        if self.total_ifb_value :
            pyk.fprint \
                ( self.out_format
                % ( self.ifb_type.name, "", "", "", self.ifb_type.abbr
                  , self.total_ifb_value
                  )
                )
    # end def write

    def _write_entry (self, e) :
        ifb_indicator = ""
        if e.ifb :
            ifb_indicator = e.ifb.abbr
        pyk.fprint \
            ( self.entry1_format
            % ( e.desc
              , ifb_indicator
              , e.birth_time.formatted ("%d.%m.%y")
              , e.birth_value
              , e.head_value
              , e.base_rate
              , e.current_depreciation
              , e.tail_value
              )
            )
        if e.alive :
            if e.ifb and e.ifb.is_new :
                pyk.fprint \
                    ( self.newifb_format
                    % ( e.supplier, "", "", "", e.ifb.abbr, e.ifb.value, "")
                    )
            elif e.ifb :
                pyk.fprint ("  %-36s%10.2f" % (e.supplier, e.ifb.value))
            else :
                pyk.fprint \
                    ( self.alive_format
                    % (e.supplier, "", "", "", ("", "ewig") ["=" in e.flags])
                    )
        else :
            pyk.fprint \
                ( self.dying_format
                % ( e.supplier
                  , "Abgang"
                  , e.death_time.formatted ("%d.%m.%y")
                  , ifb_indicator
                  , ("", e.ifb.value.as_string_s ()) [bool (e.ifb)]
                  , ("", "ewig") ["=" in e.flags]
                  , e.out_value
                  , ""
                  )
                )
    # end def _write_entry

    def update_accounts (self) :
        if self.account_file :
            file = open (self.account_file, "w")
        else :
            file = sys.stdout
        for e in self.entries :
            if e.contemporary :
                self._update_account_entry (e, file)
        if self.account_file :
            file.close ()
    # end def update_accounts

    def _update_account_entry (self, e, file) :
        cat = "fe"
        if e.p_konto :
            cat = "%sP[%s]" % (cat, e.p_konto)
        eoy = Date (day_to_time_tuple  ("31.12."))
        if e.active and e.current_depreciation :
            self._write \
                ( file
                , self.account_format
                % ( eoy.formatted ("%d.%m.")
                  , e.current_depreciation, 7800, cat, "Afa",      e.desc
                  )
                )
        if e.ifb and e.ifb.is_new and e.ifb.account :
            self._write \
                ( file
                , self.account_format
                % (eoy.formatted ("%d.%m.")
                  , e.ifb.value, e.ifb.account, cat, e.ifb.abbr, e.desc
                  )
                )
        if not e.alive :
            self._write \
                ( file
                , self.account_format
                % ( e.death_time.formatted ("%d.%m.")
                  , e.out_value, 7801, cat, "Abgang",   e.desc
                  )
                )
    # end def _update_account_entry

    def _write (self, file, s) :
        file.write (pyk.as_str (s))
    # end def _write


# end class Anlagenverzeichnis

def _main (cmd) :
    source_currency     = cmd.source_currency
    year                = cmd.year
    start               = cmd.Start_year
    file_name           = cmd.anlagenverzeichnis
    account_file        = cmd.account_file
    anlagenverzeichnis  = Anlagenverzeichnis \
        (year, start, file_name, source_currency, account_file)
    anlagenverzeichnis.evaluate ()
    anlagenverzeichnis.write    ()
    if cmd.update_accounts :
        anlagenverzeichnis.update_accounts ()
    return anlagenverzeichnis
# end def _main

_Command = TFL.CAO.Cmd \
    ( handler     = _main
    , args        =
        ( "year:S?Year of interest"
        , "anlagenverzeichnis:P?File defining depreciation data"
        )
    , min_args    = 2
    , max_args    = 2
    , opts        =
        ( "-account_file:P?Name of account file to update"
        , "-Start_year:S=1988?Skip all entries before `Start_year`"
        , "-update_accounts:B?Add depreciation entries to account file"
        , TFL.CAO.Arg.EUC_Source ()
        , TFL.CAO.Arg.EUC_Target ()
        , TFL.CAO.Opt.Output_Encoding (default = "utf-8")
        )
    , description = "Calculate depreciations for `year`"
    )

"""
year=2007 ; python -m ATAX.anlagenverzeichnis $year ~/EAR/anlagen_gewerbe.dat
"""
if __name__ == "__main__":
    _Command ()
### __END__ ATAX.anlagenverzeichnis



# -*- coding: utf-8 -*-
# Copyright (C) 2000-2015 Mag. Christian Tanzer. All rights reserved
# Glasauergasse 32, A--1130 Wien, Austria. tanzer@swing.co.at
# ****************************************************************************
# This module is part of the package ATAX.
#
# This module is licensed under the terms of the BSD 3-Clause License
# <http://www.c-tanzer.at/license/bsd_3c.html>.
# ****************************************************************************
#
#++
# Name
#    jahresabschluss
#
# Purpose
#    Sum accounting categories for one year (as required by Austrian tax law)
#
# Revision Dates
#     4-Jan-2000 (CT) Creation
#     5-Jan-2000 (CT) `gewerbeanteil' added
#     5-Jan-2000 (CT) `-plain' added
#    29-Jul-2001 (CT) Take advantage of `Command_Line.__getattr__` to access
#                     options without fuss
#    12-Feb-2002 (CT) Use `cmd.gewerbeanteil` instead of dictionary access
#     1-Dec-2002 (CT) Default of `-source_currency` and `-target_currency`
#                     changed from `ats` to `eur`
#    19-Feb-2006 (CT) Import from package _ATAX
#    11-Feb-2007 (CT) `string` functions replaced by `str` methods
#    17-Sep-2007 (CT) Modernized
#    17-Sep-2007 (CT) Use `Account.add_file`
#    17-Sep-2007 (CT) `main` refactored
#     3-Jan-2010 (CT) Use `TFL.CAO` instead of `TFL.Command_Line`
#    29-Oct-2015 (CT) Improve Python 3 compatibility
#    ««revision-date»»···
#--

from   __future__  import absolute_import
from   __future__  import division
from   __future__  import print_function
from   __future__  import unicode_literals

from _ATAX.accounting import *

class Command (Command) :

    default_categories  = "e"
    min_args            = 2

    def _create_account (self, cmd, categories, source_currency, vst_korrektur) :
        year       = cmd.argv.pop (0)
        konto_desc = Konto_Desc   (cmd.argv.pop (0))
        return T_Account \
            ( year          = year
            , konto_desc    = konto_desc
            , vst_korrektur = vst_korrektur
            )
    # end def _create_account

    def _output (self, cmd, account, categories, source_currency) :
        if cmd.plain :
            ATAX._.accounting.underlined = identity
        if not cmd.summary :
            account.print_konto_summary ()
            account.print_konten        ()
            print ("\f")
        account.print_ein_aus_rechnung  ()
        if cmd.gewerbeanteil and account.g_anteil != 0 :
            with open (cmd.gewerbeanteil, "wb") as gfile :
                gfile.write \
                    ( pyk.encoded
                        ( """$source_currency = "%s";\n"""
                        % EUC.target_currency.name
                        )
                    )
                gfile.write \
                    ( pyk.encoded
                        ( " 31.12.  &    &     & %8.2f & b  & 7001  & 2000  "
                          "& -  & e &  & Büroaufwand %s\n"
                        % (account.g_anteil, cmd.year)
                          )
                    )
    # end def _output

    def _arg_spec (self) :
        return ("year:S", "kontenplan:P", "account_file:S")
    # end def _arg_spec

    def _opt_spec (self) :
        return self.__super._opt_spec () + \
            ( "-gewerbeanteil:P?File to write gewerbe-anteil into"
            , "-plain:B?Don't underline"
            , "-summary:B"
            )
    # end def _opt_spec

# end class Command

if __name__ == "__main__":
    Command ()
### __END__ ATAX.jahresabschluss



# -*- coding: utf-8 -*-
# Copyright (C) 2002-2015 Mag. Christian Tanzer. All rights reserved
# Glasauergasse 32, A--1130 Wien, Austria. tanzer@swing.co.at
# ****************************************************************************
# This module is part of the package ATAX.
#
# This module is licensed under the terms of the BSD 3-Clause License
# <http://www.c-tanzer.at/license/bsd_3c.html>.
# ****************************************************************************
#
#++
# Name
#    income_tax_at
#
# Purpose
#    Calculate income tax according to Austrian tax law
#
# Revision Dates
#    17-Nov-2002 (CT) Creation
#    14-Jan-2004 (CT) `2005` added
#    14-Dec-2005 (CT) Type of `amount` changed from `I` to `F`
#     3-Dec-2006 (CT) Import from `_TFL`
#    17-Sep-2007 (CT) Use `EUC_Opt_SC` and `EUC_Opt_TC` instead of home-grown
#                     code
#     8-Jan-2009 (CT) Display `raw_value` of `cmd.amount`
#     9-Jan-2009 (CT) `2009` added, display `year`
#    16-Mar-2009 (CT) `main` changed to work with `Euro` using `Decimal`
#    16-Mar-2009 (CT) `tax_brackets` for `2009` changed to final version of law
#     3-Jan-2010 (CT) Use `TFL.CAO` instead of `TFL.Command_Line`
#     7-Jan-2010 (CT) Use `TFL.CAO.Arg.EUC` instead of `TFL.CAO.Arg.Money`
#    20-May-2015 (CT) Add `2016` to `tax_brackets`
#    29-Oct-2015 (CT) Improve Python 3 compatibility
#     9-Dec-2015 (CT) Fix `2016` of `tax_brackets`
#    ««revision-date»»···
#--

from   __future__  import absolute_import
from   __future__  import division
from   __future__  import print_function
from   __future__  import unicode_literals

from   _TFL.Date_Time    import *
from   _TFL.EU_Currency  import *
from   _TFL.pyk          import pyk

import _TFL.CAO

year = Date ().year

def tax_brackets (year) :
    if year < 2000 :
        return \
            ( (ATS (     50000), 0.10)
            , (ATS (    100000), 0.22)
            , (ATS (    150000), 0.32)
            , (ATS (    400000), 0.42)
            , (ATS (2000000000), 0.50)
            )
    elif year < 2002 :
        return \
            ( (ATS (     50000), 0.00)
            , (ATS (     50000), 0.21)
            , (ATS (    200000), 0.31)
            , (ATS (    400000), 0.41)
            , (ATS (2000000000), 0.50)
            )
    elif year < 2005 :
        return \
            ( (EUR (      3640), 0.00)
            , (EUR (      3630), 0.21)
            , (EUR (     14530), 0.31)
            , (EUR (     29070), 0.41)
            , (EUR (2000000000), 0.50)
            )
    elif year < 2009 :
        return \
            ( (EUR (     10000), 0.00)
            , (EUR (     15000), 0.3833)
            , (EUR (     26000), 0.436)
            , (EUR (2000000000), 0.50)
            )
    elif year < 2016 :
        return \
            ( (EUR (     11000), 0.000)
            , (EUR (     14000), 0.3635)
            , (EUR (     35000), 0.4321429)
            , (EUR (2000000000), 0.500)
            )
    else :
        return \
            ( (EUR (     11000), 0.000)
            , (EUR (      7000), 0.250)
            , (EUR (     13000), 0.350)
            , (EUR (     29000), 0.420)
            , (EUR (     30000), 0.480)
            , (EUR (    910000), 0.500)
            , (EUR (2000000000), 0.550) ### as of 2015, applies up to 2020
            )
# end def tax_brackets

def tax (amount, year = year) :
    result     = 0
    brackets   = tax_brackets (year)
    remains    = amount
    tax_chunks = []
    for chunk, rate in brackets :
        if remains <= 0.0 :
            break
        if chunk > remains :
            chunk = remains
        ### don't use `-=` here (or `chunk = remains` fails)
        remains  = remains - chunk
        tc       = chunk   * rate
        result  += tc
        tax_chunks.append ((chunk, rate, tc))
    return result, tax_chunks
# end def tax

def _main (cmd) :
    source_currency        = cmd.source_currency
    year                   = cmd.year
    amount                 = cmd.amount
    tax_amount, tax_chunks = tax (amount, year)
    if cmd.verbose :
        for c, r, t in tax_chunks :
            print \
                ( "%2d%% for %14s : %14s"
                % (r * 100, c.as_string_s (), t.as_string_s ())
                )
    f = ( "In %s, for a taxable income of %s [%s]\n"
          "    you pay a tax of %s (%5.2f%%) and get %s\n"
        )
    print \
        ( f
        % ( year, amount, cmd ["amount:raw"], tax_amount
          , (tax_amount / (amount / 100.0)).amount
          , amount - tax_amount
          )
        )
# end def _main

_Command = TFL.CAO.Cmd \
    ( handler     = _main
    , args        =
        ( TFL.CAO.Arg.EUC
            ( name        = "amount"
            , description = "Amount of taxable income"
            )
        ,
        )
    , min_args    = 1
    , max_args    = 1
    , opts        =
        ( "-verbose:B?Show chunks, too"
        , "-year:I=%s?Year of interest" % (year, )
        , TFL.CAO.Opt.EUC_Source ()
        , TFL.CAO.Opt.EUC_Target ()
        )
    , description = "Calculate income tax for `year`"
    )

if __name__ == "__main__" :
    _Command ()
### __END__ ATAX.income_tax_at



# -*- coding: utf-8 -*-
# Copyright (C) 2006-2016 Mag. Christian Tanzer. All rights reserved
# Glasauergasse 32, A--1130 Wien, Austria. tanzer@swing.co.at
# ****************************************************************************
# This module is part of the package ATAX.
#
# This module is licensed under the terms of the BSD 3-Clause License
# <http://www.c-tanzer.at/license/bsd_3c.html>.
# ****************************************************************************
#
#++
# Name
#    ATAX.__init__
#
# Purpose
#    Initialize package `ATAX`
#
# Revision Dates
#    11-Feb-2006 (CT) Creation
#    11-Oct-2016 (CT) Add `__version__`
#    ««revision-date»»···
#--

from _TFL.Package_Namespace import Package_Namespace

__version__ = "1.1.2"

ATAX = Package_Namespace ()

del Package_Namespace

### __END__ ATAX.__init__



# -*- coding: utf-8 -*-
# Copyright (C) 1999-2016 Mag. Christian Tanzer. All rights reserved
# Glasauergasse 32, A--1130 Wien, Austria. tanzer@swing.co.at
# ****************************************************************************
# This module is part of the package ATAX.
#
# This module is licensed under the terms of the BSD 3-Clause License
# <http://www.c-tanzer.at/license/bsd_3c.html>.
# ****************************************************************************
#
#++
# Name
#    ATAX.accounting
#
# Purpose
#    Provide classes for accounting (adapted to Austrian tax law)
#
# Revision Dates
#    31-Jan-1999 (CT)  Creation (factored from `umsatzsteuer.py')
#    31-Jan-1999 (CT)  Use `\s' instead of ` ' in regexes
#     7-Feb-1999 (CT)  `v' added to processing of `self.g_or_n'
#     4-Mar-1999 (CT)  `perl_dict_pat' added
#     4-Mar-1999 (CT)  `privat' added
#     7-Mar-1999 (CT)  Category `z' added for Einfuhrumsatzsteuer
#     3-Jan-2000 (CT)  `V_Account.add_lines' factored from
#                      `umsatzsteuer.umsatzsteuer'
#     4-Jan-2000 (CT)  `add_lines' moved from `V_Account' to `Account'
#     4-Jan-2000 (CT)  `eval_line' factored from `add_lines'
#     4-Jan-2000 (CT)  Parameter `categ_interest' moved from `add_lines' to
#                      `add'
#     4-Jan-2000 (CT)  `T_Account' and `Konto_Desc' added
#     5-Jan-2000 (CT)  `g_anteil' added
#     9-Jan-2000 (CT)  Erlösminderung und Ausgabenminderung considered in
#                      `V_Account', too
#     9-Jan-2000 (CT)  `_add_ausgabe', `_add_einnahme', `_add_umsatz', and
#                      `_add_vorsteuer' factored
#    24-Jan-2000 (CT)  Replace `/' by `*1.0/' to avoid integer truncation by
#                      division
#    24-Jan-2000 (CT)  `Account_Entry.time_cmp' added
#    24-Jan-2000 (CT)  `Account.add_lines': sort `entries' by `time'
#     6-Feb-2000 (CT)  `Account_Entry': handling of `s' in `cat' corrected
#                      (invert `self.netto', too)
#     6-Feb-2000 (CT)  `T_Account': for entries with accounts matching
#                      `t_konto_ignore_pat', vat is added if `u' in `cat'
#     6-Feb-2000 (CT)  `T_Account': handle `i' in `cat' correctly (must be
#                      added to both `vorsteuer' and `ust')
#     6-Nov-2000 (CT)  `#' added to `ignor_pat'
#     6-Nov-2000 (CT)  Exception handler added to `Account_Entry.__init__'
#    22-Feb-2001 (CT)  Use `raise' instead of `raise exc' for re-raise
#    26-Dec-2001 (CT)  `Konto_Desc` converted to class (and added `reverse`)
#    26-Dec-2001 (CT)  `umsatzsteuer` moved in here
#    27-Dec-2001 (CT)  `H_Account_Entry` and `H_Account` added
#    28-Dec-2001 (CT)  `*_total` added to `T_Account`
#    28-Dec-2001 (CT)  `umsatzsteuer` renamed to `add_account_file`
#    12-Feb-2002 (CT)  `print_summary` changed to use argument `round` instead
#                      of home-grown code for rounding of `vat_saldo`
#    13-Feb-2002 (CT)  `print_*` changed to include `target_currency` in output
#    13-Feb-2002 (CT)  `print_sep_line` changed to use `=` instead of `-`
#     4-Aug-2002 (CT)  `ignore` dictionary added to `Account`
#     4-Aug-2002 (CT)  `830[01][23]` removed from `t_konto_ignore_pat`
#     4-Aug-2002 (CT)  `print_ein_aus_rechnung` changed to print `saldo` and
#                      to skip accounts other than 4___ and 8___ for income
#                      and accounts other than 5___, 6___, and 7___ for
#                      expenses
#    11-Mar-2003 (CT)  `V_Account.print_summary` adapted to recent changes of
#                      Austrian tax law
#    11-Mar-2003 (CT)  Code `f` added to `g_or_n`
#    15-Jun-2003 (CT)  `Privatanteil` for VAT added
#    16-Jun-2003 (CT)  `_fix_vorsteuer_privat` fixed (all of the Privatanteil
#                      is VAT-taxed, <shucks>)
#    17-Jun-2003 (CT)  Various modernizations (`+=`)
#    17-Jun-2003 (CT)  `eust_gkonto` and `ige_gkonto` added and used
#    17-Jun-2003 (CT)  Unused ballast removed (e.g., `einnahmen_total`)
#     1-Nov-2003 (CT)  `print_summary_online` added
#    21-Dec-2003 (CT)  `H_Account.__init__` redefined to set `ignore` to `{}`
#    21-Dec-2003 (CT)  `_effective_amount` factored in `T_Account` and
#                      redefined in `H_Account`
#    21-Dec-2003 (CT)  `einnahmen_total` and friends re-introduced (is used
#                      after all by bk_abrechnung.py)
#    22-Dec-2003 (CT)  `print_konten` changed to restrict `head`s length to 64
#    28-Sep-2004 (CT)  Use `isinstance` instead of type comparison
#    24-Mar-2005 (CT)  Use `TFL.PL_Dict` and `_TFL.predicate`
#    29-Jul-2005 (CT)  `7020` added to `ausgaben_minderung_pat`
#    18-Nov-2005 (RSC) `7500` added to `ausgaben_minderung_pat`
#    11-Feb-2006 (CT)  Moved into package `ATAX`
#     1-May-2006 (CT)  `Account_Entry.__init__` changed to handle
#                      `erloes_minderung` properly
#     1-May-2006 (CT)  `Account_Entry.kontenzeile` fixed
#     8-May-2006 (CT)  Style
#     8-May-2006 (CT)  Use `TFL.defaultdict` instead of `TFL.PL_Dict`
#    15-May-2006 (CT)  `V_Account.finish` added (and then left empty, because
#                      adding correction for `privat` means major surgery)
#    11-Feb-2007 (CT)  `string` functions replaced by `str` methods
#    11-Feb-2007 (CT)  `T_Account.finish` changed to handle `vat` for
#                      `privat` correctly
#    11-Feb-2007 (CT)  `V_Account.finish` changed to consider `privat`
#    11-Feb-2007 (CT)  `T_Account._add_einnahme` and `T_Account._add_ausgabe`
#                      changed to check `entry.cat` for `u`
#    13-Jun-2007 (CT)  Argument `entry` of `V_Account._add_umsatz` made
#                      optional
#    13-Jun-2007 (CT)  `V_Account.finish` changed to consider `privat` properly
#    13-Jun-2007 (CT)  `T_Account.finish` changed to consider `privat` properly
#    13-Jun-2007 (CT)  s/_fix_vorsteuer_privat/_fix_ust_privat/g
#    13-Jun-2007 (CT)  Changed `T_Account._fix_ust_privat` to do the right
#                      thing
#    17-Sep-2007 (CT)  Function `add_account_file` replaced by method
#                      `add_file`
#    17-Sep-2007 (CT)  `Main` added
#    17-Sep-2007 (CT)  `_load_config` added
#                      * `erloes_minderung_pat` and `ausgaben_minderung_pat`
#                        moved from module-level into `Account_Entry`
#                      * CT-specific settings moved into external config file
#    17-Sep-2007 (CT)  `Account_Entry` changed to compute `minderung` based on
#                      occurrence of `~` (`erloes_minderung_pat` and
#                      `ausgaben_minderung_pat` only kept to support old input
#                      files)
#     4-Jan-2008 (CT)  `Main._load_config` changed to classmethod `load_config`
#     4-Jan-2008 (CT)  Use `Regexp` instead of `re.compile`
#    18-Feb-2008 (CT)  `kz_add` added and used
#    22-Feb-2008 (RSC) Add reverse charge
#    19-Mar-2008 (RSC) Fix issues with rev charge in Jahresabschluss
#     6-Apr-2008 (CT)  `print_summary_online` changed to only show non-zero
#                      categories
#    11-Apr-2008 (RSC) Add suffix 'r' or 'i' for rev-Charge/igE in Ust column
#                      for kontenzeile (requested by E. Pichler-Fruhstorfer)
#                      Fix gkonto description for rev. Charge (cut & paste
#                      error) in finish
#    11-Apr-2008 (RSC) Fix printing of entry if "dir" includes "~"
#    12-Aug-2008 (CT)  `p_konto` added (and `_Entry_` factored)
#     5-Sep-2008 (CT)  `_fix_ust_privat` corrected
#                      (`+=` instead of `-=` for `haben_saldo` of `ust_gkonto`)
#    18-Dec-2008 (CT)  `privat` converted to class variable to allow
#                      setting/modification in ATAX.config
#    11-Mar-2009 (CT)  Check `target_currency.to_euro_factor == 1.0` instead
#                      of comparing `target_currency` to `EU_Currency`
#    11-Mar-2009 (CT)  Use `EUR` instead of `EU_Currency` to create currency
#                      instances
#    16-Apr-2009 (CT)  `_vat_saldo` factored and changed to recompute `ust` to
#                      avoid 1-cent rounding errors
#    28-Apr-2009 (CT)  `vorsteuer_gut` added and used
#     4-May-2009 (CT)  s/vorsteuer_gut/ust_gut/
#    17-Sep-2009 (RSC) Fix Doc-String on "z" category after email-discussion
#     6-Dec-2009 (CT)  3-compatibility
#     3-Jan-2010 (CT) Use `TFL.CAO` instead of `TFL.Command_Line`
#    27-Jan-2010 (CT) Option `-stdin` added
#    14-Apr-2010 (CT) `Account_Entry.__init__` changed to swap `konto` and
#                     `gegen_konto` if `minderung`
#    14-Apr-2010 (CT) `T_Account.add` changed not to consider `minderung`,
#                     i.e., to always call `_add_ausgabe` for `-`,
#                     `_add_einnahme` for `+`
#     4-Jun-2012 (MG) `_ATAX_Command_._add_files` support for glob patterns
#                     added
#     7-Jun-2012 (CT) Use `TFL.r_eval`
#     2-Jan-2014 (CT) Add and use `_decoded`
#     3-Jan-2014 (CT) Add `output_encoding`; use `pyk.fprint`, not `print`
#    14-Aug-2014 (CT) Add comment to clause for `ausgaben_minderung`
#     4-Dec-2014 (MG) Add `T_Account.calculation_method` to allow
#                     customization of calculation method
#     4-Dec-2014 (CT) Remove debug output for `calculation_method`
#    17-Feb-2015 (CT) Move values for `ignore` to `Config`, out of `__init__`
#     2-Jun-2015 (CT) Add dictionary `vat_privat` indexed by account,
#                     rename `vat_private` to `vat_privat_default`
#     8-Oct-2015 (CT) Change `__getattr__` to *not* handle `__XXX__`
#    29-Oct-2015 (CT) Improve Python 3 compatibility
#    29-Oct-2015 (CT) Fix usage of `vat_privat` (use `get`, not `[]`)
#     9-Oct-2016 (CT) Remove dependency on `TGL`
#    ««revision-date»»···
#--

from   __future__  import absolute_import
from   __future__  import division
from   __future__  import print_function
from   __future__  import unicode_literals

from   _ATAX             import ATAX
from   _TFL              import TFL

from   _TFL.Date_Time    import *
from   _TFL.EU_Currency  import *
from   _TFL.defaultdict  import defaultdict
from   _TFL.predicate    import *
from   _TFL.pyk          import pyk
from   _TFL.Regexp       import *

import _TFL._Meta.Object
import _TFL.Accessor
import _TFL.CAO
import _TFL.r_eval

import _TFL.load_config_file

import math
import sys
import glob

ignor_pat              = Regexp ( r"^\s*[«%#]")
empty_pat              = Regexp ( r"^\s*$")
ws_head_pat            = Regexp ( r"^\s*")
ws_tail_pat            = Regexp ( r"\s*\n?$")
code_pat               = Regexp ( r"^\s*\$")
perl_dict_pat          = Regexp ( r"""\{\s*"?(\s*\d+\s*)"?\s*\}""")
split_pat              = Regexp ( r"\s*&\s*")
currency_pat           = Regexp ( r"([A-Za-z]+)$")
desc_strip_pat         = Regexp ( r"\s*&\s*$")

def underlined (text) :
    bu = "\b_"
    return bu.join (text) + bu
# end def underlined

class _Base_ (TFL.Meta.Object) :

    def _decoded (self, line) :
        try :
            result = line.decode ("utf-8")
        except Exception as exc :
            result = line.decode ("iso-8859-1")
        return result
    # end def _decoded

# end class _Base_

class _Entry_ (_Base_) :

    _p_konto_pat = Regexp (r"P\[(?P<p_konto> \d+)\]", re.VERBOSE)

    def _get_p_konto (self, flags) :
        pat = self._p_konto_pat
        if pat.search (flags) :
            return pat.p_konto
    # end def _get_p_konto

# end class _Entry_

class Account_Entry (_Entry_) :
    """Entry of accounting file."""

    """Cat (several letters can be combined)::

            e       include in income tax calculations
            f       financial planning
            i       Innergemeinschaftlicher Erwerb
            k       correction
            r       reverse charge (for VAT, Article 19 [057/066 in VAT form])
            s       storno
            u       include in VAT calculations
            z       pure VAT amount (Einfuhrumsatzsteuer)

        g_or_n (exactly one letter)::

            b       gross amount
            f       VAT free
            n       net amount
            v       pure VAT amount
    """

    ### The following class variables can be overriden in a config file
    ### (e.g., ATAX.config)
    erloes_minderung_pat   = Regexp ( r"Legacy: use `~` instead")
    ausgaben_minderung_pat = Regexp ( r"Legacy: use `~` instead")

    minderung              = False

    def __init__ (self, line, source_currency, vst_korrektur = 1.0) :
        self.line = line
        try :
            ( self.date, self.number, self.vat_txt, gross,    self.g_or_n
            , self.soll, self.haben,  self.dir,     self.cat, self.plan_kat
            , desc
            )               = split_pat.split    (line, 10)
            self.dtuple     = day_to_time_tuple  (self.date)
        except (ValueError, TypeError) as exc :
            pyk.fprint (exc)
            pyk.fprint (line)
            raise
        self.desc           = desc_strip_pat.sub ("", desc)
        self.vst_korrektur  = vst_korrektur
        self.time           = mktime             (tuple (self.dtuple))
        self.p_konto        = self._get_p_konto  (self.cat)
        self.vat_type       = ' '
        for k in 'ir' :
            if k in self.cat :
                self.vat_type = k
                break
        currency_match      = currency_pat.search (gross)
        if currency_match :
            source_currency = EUR.Table [currency_match.group (1)]
            gross           = currency_pat.sub ("", gross)
        ### avoid integer truncation by division
        ### otherwise `1/2' evaluates to `0.0' :-(
        gross      = gross.replace   ("/", "*1.0/")
        ###
        self.gross = source_currency (TFL.r_eval (gross))
        self.netto = source_currency (self.gross)
        if   "s" in self.cat :
            self.gross = - self.gross
            self.netto = - self.netto
            self.flag  = "S"
        elif "k" in self.cat :
            self.flag  = "k"
        else :
            self.flag  = " "
        self.vat_p = (TFL.r_eval (self.vat_txt or "0") / 100.0) + 1.0
        if   "n" == self.g_or_n :
            self.gross = self.netto * self.vat_p
        elif "b" == self.g_or_n :
            self.netto = self.gross / self.vat_p
        elif "v" == self.g_or_n :
            ### entry is pure VAT amount (Einfuhr-Umsatzsteuer)
            self.netto = source_currency (0)
        self.vat       = self.gross - self.netto
        self.vat_vstk  = source_currency (0)
        self.vat_x     = source_currency (0)
        self.is_change = 1
        self.minderung = "~" in self.dir
        if   "-" in self.dir :
            self.soll_betrag   = self.netto
            self.haben_betrag  = source_currency (0)
            self.minderung     = \
                self.minderung or self.erloes_minderung_pat.match (self.haben)
            if not self.minderung :
                self.konto         = self.soll
                self.gegen_konto   = self.haben
                if self.vat_p != 1 :
                    self.vat_vstk  = self.vat
                    self.vat       = self.vat * vst_korrektur
                    self.vat_vstk -= self.vat
                    self.vat_x    = self.vat
            else :
                self.konto         = self.haben
                self.gegen_konto   = self.soll
        elif "+" in self.dir :
            self.haben_betrag  = self.netto
            self.soll_betrag   = source_currency (0)
            self.minderung     = \
                (  self.minderung
                or self.ausgaben_minderung_pat.match (self.haben)
                )
            if not self.minderung :
                self.konto         = self.haben
                self.gegen_konto   = self.soll
            else :
                ### In many cases, it is better to book dir `-` with a
                ### negative amount instead
                ###
                ### XXX VAT should be handled properly here: compare `-`
                self.konto         = self.soll
                self.gegen_konto   = self.haben
        else :
            self.is_change     = 0
        self.cati = " "
    # end def __init__

    def __getattr__ (self, name) :
        if name.startswith ("__") and name.endswith ("__") :
            ### Placate inspect.unwrap of Python 3.5,
            ### which accesses `__wrapped__` and eventually throws `ValueError`
            return getattr (self.__super, name)
        return getattr (self.dtuple, name)
    # end def __getattr__

    def __str__ (self) :
        return "%6s %2.2f%s%10s %10s %10s  %-5s  %-5s%2s%1s %s" % \
            ( self.date, self.vat_p, self.cati
            , self.vat.as_string   ()
            , self.netto.as_string ()
            , self.gross.as_string ()
            , self.soll, self.haben, self.dir
            , self.flag, self.desc [:18]
            )
    # end def __str__

    def kontenzeile (self) :
        ## print self.vat_p, type (self.vat_p) # self.soll_betrag, self.haben_betrag
        try :
            vat_type = getattr (self, "vat_type", " ")
            return "%02d%02d  %-5s  %-35.35s %2s%1s %12s  %12s" % \
                ( self.day, self.month
                , self.gegen_konto
                , self.desc
                , self.vat_txt
                , vat_type
                , self._soll_betrag  ()
                , self._haben_betrag ()
                )
        except Exception as exc :
            pyk.fprint \
              ( exc
              , type (self.soll_betrag),  self.soll_betrag
              , type (self.haben_betrag), self.haben_betrag
              )
            ### raise
    # end def kontenzeile

    def _soll_betrag (self) :
        if self.soll_betrag != 0 or "-" in self.dir :
            return self.soll_betrag.as_string_s ()
        else :
            ### if "+" not in self.dir : return "<< %s >>" % self.dir
            return ""
    # end def _soll_betrag

    def _haben_betrag (self) :
        if self.haben_betrag != 0 or "+" in self.dir :
            return self.haben_betrag.as_string_s ()
        else :
            ### if "-" not in self.dir : return "<< %s >>" % self.dir
            return ""
    # end def _haben_betrag

    def _init (self, soll_betrag, haben_betrag, gegen_konto = "3293", konto = "") :
        self.konto        = konto
        self.soll_betrag  = EUR (soll_betrag)
        self.haben_betrag = EUR (haben_betrag)
        self.dtuple       = day_to_time_tuple  (self.date)
        self.gegen_konto  = gegen_konto
        self.vat_txt      = self.vat_p = self.cat_i = ""
        self.soll         = self.haben = self.flag  = self.dir = ""
        self.gross        = self.netto = self.vat = EUR (0)
    # end def _init

# end class Account_Entry

class Ust_Gegenbuchung (Account_Entry) :

    def __init__ (self, month, konto, soll_betrag, haben_betrag, desc) :
        self.date = "%s.%s" % (Time_Tuple.day_per_month [month - 1], month)
        self._init  (soll_betrag, haben_betrag, konto = konto)
        self.desc = "%-12s %02d" % (desc, self.dtuple.month)
    # end def __init__

# end class Ust_Gegenbuchung

class Privatanteil (Account_Entry) :

    def __init__ (self, gegen_konto, soll_betrag, haben_betrag, desc, konto = "") :
        self.date = "31.12."
        self._init  (soll_betrag, haben_betrag, gegen_konto, konto)
        self.desc = "%-12s" % (desc, )
    # end def __init__

# end class Privatanteil

class Account (_Base_):
    """Account for a specific time interval (e.g., month, quarter, or year)"""

    Entry          = Account_Entry

    ### The following class variables can be overriden in a config file
    ### (e.g., ATAX.config)
    ignore             = ()
    gewerbe_anteil     = 0
    privat             = {}
    vat_privat         = {}   ### VAT rate per account in `.privat`
    vat_privat_default = 1.20 ### default VAT rate applicable for private part

    def __init__ (self, name = "", vst_korrektur = 1.0) :
        self.name           = name
        self.vst_korrektur  = vst_korrektur
        self.entries        = []
        self.privat         = dict (self.privat)
        self.vat_privat     = dict (self.vat_privat)
        self.ignore         = set  (self.__class__.ignore)
        self._finished      = False
    # end def __init__

    account_vars = set (("vst_korrektur", "firma", "gewerbe_anteil"))

    def add_file (self, file_name, categ_interest, source_currency) :
        file = open (file_name, "rb")
        try :
            self.add_lines (file, categ_interest, source_currency)
        finally :
            file.close ()
    # end def add_file

    def add_lines (self, lines, categ_interest, source_currency) :
        """Add entries given by `lines' to account `self'."""
        self.source_currency = source_currency
        entries = []
        for line in lines :
            line                      = self._decoded (line)
            if ignor_pat.match (line) or empty_pat.match (line) :
                continue
            line                      = ws_head_pat.sub ("", line, count = 1)
            line                      = ws_tail_pat.sub ("", line, count = 1)
            if not line               : continue
            if code_pat.match  (line) :
                self.eval_line (line)
            else :
                entry = self.Entry \
                    (line, self.source_currency, self.vst_korrektur)
                if (categ_interest.search (entry.cat) and entry.is_change) :
                    entries.append (entry)
        entries.sort (key = TFL.Getter.time)
        self.entries.extend (entries)
        for entry in entries :
            self.add (entry)
    # end def add_lines

    def eval_line (self, line) :
        line = code_pat.sub ("", line, count = 1)
        if perl_dict_pat.search (line) :
            line = perl_dict_pat.sub ("""["\\1"]""", line)
        ### use temporary dict for exec
        ### (`locals ()' cannot be changed by `exec')
        tmp  = { "privat" : {}, "vat_privat" : {}, "ignore" : {}}
        try :
            ### `tmp' must be passed to the local-dict argument because
            ### Python adds some junk to the global-dict argument of `exec'
            exec (line, {}, tmp)
            if "source_currency" in tmp :
                self.source_currency = EUC.Table [tmp ["source_currency"]]
                del tmp ["source_currency"]
            self.privat.update     (tmp ["privat"])
            self.vat_privat.update (tmp ["vat_privat"])
            self.ignore.update     (tmp ["ignore"])
            del tmp ["privat"]
            del tmp ["vat_privat"]
            del tmp ["ignore"]
            try :
                for k, v in tmp.items () :
                    if k in self.account_vars :
                        setattr (self, k, v)
                    else :
                        pyk.fprint ("Invalid assignment `%s'" % line)
            except Exception as exc :
                pyk.fprint \
                    ( "Exception `%s' during local-dict update `%s'"
                    % (exc, tmp_cmd)
                    )
        except Exception as exc :
            pyk.fprint \
                ( "Exception `%s' encountered during execution of line"
                  "\n   `%s'"
                % (exc, line)
                )
    # end def eval_line

# end class Account

class V_Account (Account) :
    """VAT Account for a specific time interval (e.g., month, quarter, or year)"""

    ### The following class variables can be overriden in a config file
    ### (e.g., ATAX.config)
    kz_add      = \
        { "????"  : ("027", "In 060/065 enthaltene Vorsteuern betreffend KFZ")
        }

    def __init__ (self, name = "", vst_korrektur = 1.0) :
        Account.__init__ (self, name, vst_korrektur)
        self.ausgaben_b              = EUR (0)
        self.ausgaben_n              = EUR (0)
        self.erwerb_igE              = EUR (0)
        self.reverse_charge          = EUR (0)
        self.umsatz                  = EUR (0)
        self.umsatz_frei             = EUR (0)
        self.ust                     = EUR (0)
        self.ust_igE                 = EUR (0)
        self.ust_revCharge           = EUR (0)
        self.vorsteuer               = EUR (0)
        self.ust_gut                 = EUR (0)
        self.vorsteuer_igE           = EUR (0)
        self.vorsteuer_revCh         = EUR (0)
        self.vorsteuer_EUst          = EUR (0)
        self.erwerb_igE_dict         = defaultdict (EUR)
        self.reverse_charge_dict     = defaultdict (EUR)
        self.umsatz_dict             = defaultdict (EUR)
        self.ust_dict                = defaultdict (EUR)
        self.ust_igE_dict            = defaultdict (EUR)
        self.ust_revC_dict           = defaultdict (EUR)
        self.vorsteuer_kzs           = defaultdict (EUR)
        self.umsatzsteuer_entries    = []
        self.vorsteuer_entries       = []
        self.vorsteuer_entries_igE   = []
        self.vorsteuer_entries_revC  = []
        self.vorsteuer_entries_EUst  = []
    # end def __init__

    def add (self, entry) :
        """Add `entry' to U_Account."""
        assert (isinstance (entry, Account_Entry))
        vst_korrektur = entry.vst_korrektur
        netto         = entry.netto
        gross         = entry.gross
        vat           = entry.vat
        vat_p         = entry.vat_p
        if   "-" in entry.dir :
            self.ausgaben_b  += gross
            self.ausgaben_n  += netto
            if "i" in entry.cat :
                self.vorsteuer_igE += vat
                self.ust_igE       += vat
                self.vorsteuer_entries_igE.append (entry)
                entry.cati = "i"
                self.ust_igE_dict  [vat_p]   += vat
                self.erwerb_igE              += netto
                self.erwerb_igE_dict [vat_p] += netto
                if vst_korrektur != 1.0 :
                    sys.stderr.write \
                        ( "Cannot handle expenses resulting from "
                          "innergemeinschaftlichem Erwerb for a "
                          "VAT correction of %d\n       %s\n"
                        % (vst_korrektur, entry)
                        )
                elif "b" == entry.g_or_n :
                    sys.stderr.write \
                        ( "**** igE entries must be netto****\n       %s\n"
                        % entry
                        )
                if entry.konto in self.kz_add :
                    self.vorsteuer_kzs [self.kz_add [entry.konto]] += vat
            elif "r" in entry.cat :
                self.vorsteuer_revCh     += vat
                self.ust_revCharge       += vat
                self.vorsteuer_entries_revC.append (entry)
                entry.cati = "r"
                self.ust_revC_dict  [vat_p]      += vat
                self.reverse_charge              += netto
                self.reverse_charge_dict [vat_p] += netto
                if vst_korrektur != 1.0 :
                    sys.stderr.write \
                        ( "Cannot handle expenses resulting from "
                          "reverse Charge for a "
                          "VAT correction of %d\n       %s\n"
                        % (vst_korrektur, entry)
                        )
                elif "b" == entry.g_or_n :
                    sys.stderr.write \
                        ( "**** reverse Charge entries must be netto****\n"
                          "       %s\n"
                        % entry
                        )
                if entry.konto in self.kz_add :
                    self.vorsteuer_kzs [self.kz_add [entry.konto]] += vat
            elif "z" in entry.cat :
                self.vorsteuer_EUst += vat
                self.vorsteuer_entries_EUst.append (entry)
                entry.cati = "E"
                if vst_korrektur != 1.0 :
                    sys.stderr.write \
                        ( "Cannot handle Einfuhrumsatzsteuer for a "
                          "VAT correction of %d\n       %s\n"
                        % (vst_korrektur, entry)
                        )
            else : ### neither "i" nor "z" nor "r"
                self.vorsteuer_entries.append (entry)
                if entry.minderung :
                    ## Erlösminderung instead of Ausgabe
                    self._add_umsatz    (- netto, - vat, vat_p, entry)
                else :
                    self._add_vorsteuer (vat)
                    if entry.konto in self.kz_add :
                        self.vorsteuer_kzs [self.kz_add [entry.konto]] += vat
        elif "+" in entry.dir :
            self.umsatzsteuer_entries.append (entry)
            if "i" in entry.cat :
                sys.stderr.write \
                    ( "Cannot handle income resulting from "
                      "innergemeinschaftlichem Erwerb\n       %s\n"
                    % entry
                    )
            if "r" in entry.cat :
                sys.stderr.write \
                    ( "Cannot handle income resulting from "
                      "reverse Charge\n       %s\n"
                    % entry
                    )
            if entry.minderung :
                ## Ausgabenminderung instead of Einnahme
                self.ust_gut += vat
            else :
                self._add_umsatz (netto, vat, vat_p, entry)
    # end def add

    def _add_umsatz (self, netto, vat, vat_p, entry = None) :
        self.ust                  += vat
        self.umsatz               += netto
        self.umsatz_dict [vat_p]  += netto
        if entry and entry.g_or_n == "f" :
            self.umsatz_frei      += netto
        elif vat_p != 1.0 :
            self.ust_dict [vat_p] += vat
    # end def _add_umsatz

    def _add_vorsteuer (self, vat) :
        self.vorsteuer += vat
    # end def _add_vorsteuer

    def finish (self) :
        if not self._finished :
            self._finished = True
            netto  = defaultdict (EUR)
            p_vat  = defaultdict (EUR)
            vat_p  = self.vat_privat
            vat_pd = self.vat_privat_default
            for entry in self.vorsteuer_entries :
                k = entry.konto
                if k in self.privat and entry.netto > 0 :
                    vat_p.setdefault (k, vat_pd)
                    factor      = self.privat [k] / 100.0
                    corr        = entry.netto * factor
                    netto  [k] += corr
                    p_vat  [k] += corr * (vat_p [k] - 1.00)
            for k, n in pyk.iteritems (netto) :
                if n :
                    self._add_umsatz (n, p_vat [k], vat_p [k])
    # end def finish

    def header_string (self) :
        return " %s   %8s%7s   %8s   %8s  %-5s %-5s %s  %s\n%s" % \
            ( "Tag"
            ,      "MSt-Satz"
            ,         "MWSt"
            ,               "Netto"
            ,                     "Brutto"
            ,                          "Soll"
            ,                               "Haben"
            ,                                    "  "
            ,                                        "Text"
            ,                                            "=" * 80
            )
    # end def header_string

    def print_entries (self, trailer = None) :
        """Print `self.entries' followed by trailer"""
        self.print_entries_ (self.entries, self.header_string (), trailer)
    # end def print_entries

    def print_entries_by_group (self, trailer = None) :
        """Print `self.vorsteuer_entries', `self.vorsteuer_entries_igE', and
           `self.umsatzsteuer_entries' followed by trailer.
        """
        if  (  self.vorsteuer_entries
            or self.vorsteuer_entries_EUst
            or self.vorsteuer_entries_igE
            or self.vorsteuer_entries_revC
            or self.umsatzsteuer_entries
            ) :
            pyk.fprint (self.header_string ())
        self.print_entries_ (self.vorsteuer_entries,      trailer = "\n")
        self.print_entries_ (self.vorsteuer_entries_EUst, trailer = "\n")
        self.print_entries_ (self.vorsteuer_entries_igE,  trailer = "\n")
        self.print_entries_ (self.vorsteuer_entries_revC, trailer = "\n")
        self.print_entries_ (self.umsatzsteuer_entries)
        if  (  self.vorsteuer_entries
            or self.vorsteuer_entries_EUst
            or self.vorsteuer_entries_igE
            or self.vorsteuer_entries_revC
            or self.umsatzsteuer_entries
            ) :
            if trailer :
                pyk.fprint (trailer)
    # end def print_entries_by_group

    def print_entries_ (self, entries, header = None, trailer = None) :
        if entries :
            if header :
                pyk.fprint (header)
            for e in entries :
                pyk.fprint (e)
            if trailer :
                pyk.fprint (trailer)
    # end def print_entries_

    def print_summary_old (self) :
        """Print summary for Umsatzsteuervoranmeldung."""
        vat_saldo = self._vat_saldo ()
        meldung   = ("Überschuss", "Vorauszahlung") [vat_saldo >= 0]
        sign      = (-1.0,         +1.0)            [vat_saldo >= 0]
        pyk.fprint \
            ( "%-16s : %14s"
            % ("Ausgaben brutto", self.ausgaben_b.as_string_s ())
            )
        pyk.fprint \
            ( "%-16s : %14s"
            % ("Ausgaben netto",  self.ausgaben_n.as_string_s ())
            )
        pyk.fprint ("\n%s\n"         % ( "=" * 80, ))
        pyk.fprint ("%-16s : %14s"   % ( "Umsatz", self.umsatz.as_string_s ()))
        pyk.fprint \
            ( "%-16s : %14s"
            %  ( "Steuerpflichtig"
               , (self.umsatz - self.umsatz_dict [1.0]).as_string_s ()
               )
            )
        self.print_ust_dict_     ( self.umsatz_dict, self.ust_dict)
        pyk.fprint \
            ( "\n%-16s : %14s"
            % ("Erwerbe igE", self.erwerb_igE.as_string_s ())
            )
        self.print_ust_dict_     ( self.erwerb_igE_dict, self.ust_igE_dict)
        pyk.fprint \
            ( "\n%-16s : %14s"
            % ("Reverse Charge", self.reverse_charge.as_string_s ())
            )
        self.print_ust_dict_     ( self.reverse_charge_dict, self.ust_revC_dict)
        pyk.fprint \
            ( "\n%-16s :                %14s"
            % ( "USt Übertrag"
              , (self.ust + self.ust_igE + self.ust_revCharge).as_string_s ()
              )
            )
        pyk.fprint ("--------------- ./.. ---------------------------")
        pyk.fprint \
            ( "%-16s :                %14s"
            % ( "USt Übertrag"
              , (self.ust + self.ust_igE + self.ust_revCharge).as_string_s ()
              )
            )
        pyk.fprint \
            ( "%-16s : %14s"
            % ( "Vorsteuer", self.vorsteuer.as_string_s ())
            )
        pyk.fprint \
            ( "%-16s : %14s"
            % ( "Umsatzsteuer aus Gutschrift", self.ust_gut.as_string_s ())
            )
        pyk.fprint \
            ( "%-16s : %14s"
            % ("Einfuhrumsatzst.", self.vorsteuer_EUst.as_string_s ())
            )
        pyk.fprint \
            ( "%-16s : %14s"
            % ( "Vorsteuer igE", self.vorsteuer_igE.as_string_s ())
            )
        pyk.fprint \
            ( "%-16s : %14s %14s"
            % ( "Summe Vorsteuer"
              , ( self.vorsteuer     - self.ust_gut + self.vorsteuer_EUst
                + self.vorsteuer_igE + self.vorsteuer_revCh
                ).as_string_s ()
              , ( self.vorsteuer     - self.ust_gut + self.vorsteuer_EUst
                + self.vorsteuer_igE + self.vorsteuer_revCh
                ).as_string_s ()
              )
            )
        if vat_saldo.target_currency.to_euro_factor == 1.0 :
            ### no rounding for Euro
            pyk.fprint \
                ( "\n%-16s :                %14s %s"
                % ( meldung
                  , vat_saldo.as_string_s (), vat_saldo.target_currency.name
                  )
                )
        else :
            ### rounding is necessary
            pyk.fprint \
                ( "\n%-16s :             %14s%s00 %s     (%s)"
                % ( meldung
                  , vat_saldo.as_string_s (round = 1)
                  , vat_saldo.target_currency.decimal_sign
                  , vat_saldo.target_currency.name
                  , vat_saldo.as_string_s ()
                  )
                )
    # end def print_summary_old

    def print_summary (self) :
        """Print summary for Umsatzsteuervoranmeldung."""
        vat_saldo    = self._vat_saldo ()
        meldung      = ("Überschuss", "Vorauszahlung") [vat_saldo >= 0]
        sign         = (-1.0,         +1.0)            [vat_saldo >= 0]
        umsatz_vat   = self.umsatz - self.umsatz_frei
        pyk.fprint \
            ( "%-30s     : %29s"
            % ("Ausgaben brutto", self.ausgaben_b.as_string_s ())
            )
        pyk.fprint \
            ( "%-30s     : %29s"
            % ("Ausgaben netto", self.ausgaben_n.as_string_s ())
            )
        pyk.fprint ("\n%s\n"           % ( "=" * 80, ))
        pyk.fprint \
            ( "%-30s %3s : %29s"
            % ("Nichtsteuerbar Ausland", "005", self.umsatz_frei.as_string_s())
            )
        pyk.fprint \
            ( "%-30s %3s : %29s"
            % ("Lieferungen, sonstige", "000", umsatz_vat.as_string_s ())
            )
        pyk.fprint \
            ( "%-30s     : %29s"
            % ("Summe Bemessungsgrundlage", umsatz_vat.as_string_s ())
            )
        pyk.fprint ()
        pyk.fprint \
            ( "%-30s     : %29s"
            % ("Gesamt steuerpflichtig", umsatz_vat.as_string_s ())
            )
        self.print_ust_dict_  (self.umsatz_dict, self.ust_dict, self._ust_cat)
        pyk.fprint \
            ( "%-30s %3s : %29s"
            % ( "Reverse Charge §19", "057"
              , self.vorsteuer_revCh.as_string_s ()
              )
            )
        pyk.fprint ()
        pyk.fprint \
            ("%-30s     : %29s" % ( "USt Übertrag", self.ust.as_string_s ()))
        pyk.fprint ("-"*33, "./..", "-" * 27)
        pyk.fprint \
            ("%-30s     : %29s" % ( "USt Übertrag", self.ust.as_string_s ()))
        pyk.fprint \
            ( "%-30s %3s : %14s"
            % ("Erwerbe igE", "070", self.erwerb_igE.as_string_s ())
            )
        self.print_ust_dict_ \
            (self.erwerb_igE_dict, self.ust_igE_dict, self._ige_cat)
        pyk.fprint \
            ( "%-30s %3s : %29s"
            % ("Vorsteuer", "060", self.vorsteuer.as_string_s ())
            )
        pyk.fprint \
            ( "%-30s %3s : %29s"
            % ("Einfuhrumsatzsteuer", "061", self.vorsteuer_EUst.as_string_s ())
            )
        pyk.fprint \
            ( "%-30s %3s : %29s"
            % ("Vorsteuer igE", "065", self.vorsteuer_igE.as_string_s ())
            )
        pyk.fprint \
            ( "%-30s %3s : %29s"
            % ( "Reverse Charge §19", "066"
              , self.vorsteuer_revCh.as_string_s ()
              )
            )
        for (k, d), vst in sorted (pyk.iteritems (self.vorsteuer_kzs)) :
            pyk.fprint ("%-30.30s %3s : %29s" % (d, k, vst.as_string_s ()))
        pyk.fprint \
            ( "%-30s %3s : %29s"
            % ( "Umsatzsteuer aus Gutschrift", "090"
              , self.ust_gut.as_string_s ()
              )
            )
        pyk.fprint \
            ( "%-30s     : %29s"
            % ( "Gesamtbetrag Vorsteuer"
              , ( self.vorsteuer     - self.ust_gut + self.vorsteuer_EUst
                + self.vorsteuer_igE + self.vorsteuer_revCh
                ).as_string_s ()
              )
            )
        pyk.fprint ()
        if vat_saldo.target_currency.to_euro_factor == 1.0 :
            ### no rounding for Euro
            pyk.fprint \
                ( "%-30s %3s : %29s %s"
                % ( meldung, "095"
                  , vat_saldo.as_string_s (), vat_saldo.target_currency.name
                  )
                )
        else :
            ### rounding is necessary
            pyk.fprint \
                ( "%-30s %3s : %29s%s00 %s     (%s)"
                % ( meldung, "095"
                  , vat_saldo.as_string_s (round = 1)
                  , vat_saldo.target_currency.decimal_sign
                  , vat_saldo.target_currency.name
                  , vat_saldo.as_string_s ()
                  )
                )
    # end def print_summary

    def print_summary_online (self) :
        """Print summary for Umsatzsteuervoranmeldung for online entry."""
        vat_saldo    = self._vat_saldo ()
        meldung      = ("Überschuss", "Vorauszahlung") [vat_saldo >= 0]
        sign         = (-1.0,         +1.0)            [vat_saldo >= 0]
        umsatz_vat   = self.umsatz - self.umsatz_frei
        pyk.fprint \
            ("*** Lieferungen, sonstige Leistungen und Eigenverbrauch ***")
        pyk.fprint ("=" * 67)
        if umsatz_vat :
            pyk.fprint \
                ( "%-50s %3s : %10s"
                % ("Lieferungen, sonstige", "000", umsatz_vat.as_string_s ())
                )
        if self.umsatz_frei :
            pyk.fprint \
                ( "%-50s %3s : %10s"
                % ( "Nichtsteuerbar Ausland", "005"
                  , self.umsatz_frei.as_string_s ()
                  )
                )
        if umsatz_vat :
            pyk.fprint ()
            pyk.fprint ("Steuersätze")
            pyk.fprint ("=" * 67)
            self.print_ust_dict_online (self.umsatz_dict, self._ust_cat)
        if self.vorsteuer_revCh :
            pyk.fprint \
                ( "\n%-50s %3s : %10s"
                % ( "Reverse Charge §19", "057"
                  , self.vorsteuer_revCh.as_string_s ()
                  )
                )
        pyk.fprint ("\n\n")
        pyk.fprint ("*** Innergemeinschaftliche Erwerbe ***")
        pyk.fprint ("=" * 67)
        if self.erwerb_igE :
            pyk.fprint \
                ( "%-50s %3s : %10s"
                % ("Erwerbe igE", "070", self.erwerb_igE.as_string_s ())
                )
            pyk.fprint ()
            pyk.fprint ("Steuersätze")
            pyk.fprint ("=" * 47)
            self.print_ust_dict_online (self.erwerb_igE_dict, self._ige_cat)
        pyk.fprint ("\n\n")
        pyk.fprint ("*** Vorsteuer ***")
        pyk.fprint ("=" * 67)
        pyk.fprint \
            ( "%-50s %3s : %10s"
            % ("Vorsteuer", "060", self.vorsteuer.as_string_s ())
            )
        if self.vorsteuer_EUst:
            pyk.fprint \
                ( "%-50s %3s : %10s"
                % ( "Einfuhrumsatzsteuer", "061"
                  , self.vorsteuer_EUst.as_string_s ()
                  )
                )
        if self.vorsteuer_igE :
            pyk.fprint \
                ( "%-50s %3s : %10s"
                % ("Vorsteuer igE", "065", self.vorsteuer_igE.as_string_s ())
                )
        if self.vorsteuer_revCh :
            pyk.fprint \
                ( "%-50s %3s : %10s"
                % ( "Reverse Charge §19", "066"
                  , self.vorsteuer_revCh.as_string_s ()
                  )
                )
        for (k, d), vst in sorted (pyk.iteritems (self.vorsteuer_kzs)) :
            pyk.fprint ("%-50.50s %3s : %10s" % (d, k, vst.as_string_s ()))
        if self.ust_gut :
            pyk.fprint \
                ( "%-50s %3s : %10s"
                % ( "Umsatzsteuer auf Gutschrift", "090"
                  , self.ust_gut.as_string_s ()
                  )
                )
        pyk.fprint ("\n\n")
        pyk.fprint ("*" * 67)
        if vat_saldo.target_currency.to_euro_factor == 1.0 :
            ### no rounding for Euro
            pyk.fprint \
                ( "%-50s %3s : %10s %s"
                % ( meldung, "095"
                  , vat_saldo.as_string_s (), vat_saldo.target_currency.name
                  )
                )
        else :
            ### rounding is necessary
            pyk.fprint \
                ( "%-50s %3s : %10s%s00 %s     (%s)"
                % ( meldung, "095"
                  , vat_saldo.as_string_s (round = 1)
                  , vat_saldo.target_currency.decimal_sign
                  , vat_saldo.target_currency.name
                  , vat_saldo.as_string_s ()
                  )
                )
    # end def print_summary_online

    _ust_cat = {20 : "022", 10 : "029", 0 : ""}
    _ige_cat = {20 : "072", 10 : "073", 0 : ""}

    def print_ust_dict_ (self, umsatz_dict, ust_dict, cat) :
        for vat_p in sorted (umsatz_dict, reverse = True) :
            vp = int (((vat_p - 1) * 100) + 0.5)
            pyk.fprint \
                ( "davon %2d%%                      %3s : %14s %14s"
                % ( vp, cat [vp]
                  , umsatz_dict [vat_p].as_string_s ()
                  , ust_dict    [vat_p].as_string_s ()
                  )
                )
    # end def print_ust_dict_

    def print_ust_dict_online (self, umsatz_dict, cat) :
        for vat_p in sorted (umsatz_dict, reverse = True) :
            ust = umsatz_dict [vat_p]
            if ust :
                vp = int (((vat_p - 1) * 100) + 0.5)
                hd = "davon %2d%%" % (vp, )
                pyk.fprint \
                    ("%-50s %3s : %10s" % (hd, cat [vp], ust.as_string_s ()))
    # end def print_ust_dict_online

    def _vat_saldo (self) :
        ### recompute `ust` instead of using `self.ust` to avoid 1-cent
        ### rounding errors
        ust = sum \
            ( ((v*r - v) for (r, v) in pyk.iteritems (self.umsatz_dict))
            , EUR (0)
            )
        return ust + self.ust_gut - self.vorsteuer - self.vorsteuer_EUst
    # end def _vat_saldo

# end class U_Account

class T_Account (Account) :
    """Account total for a specific time interval (e.g., month, quarter, or year.)"""

    ### The following class variables can be overriden in a config file
    ### (e.g., ATAX.config)
    eust_gkonto        = "9997"
    ige_gkonto         = "9998"
    rvc_gkonto         = "9996"
    ust_gkonto         = "9999"
    vorsteuer_gkonto   = "9999"
    t_konto_ignore_pat = Regexp (r"^[01239]\d\d\d\d?", re.X)
    firma              = "<<<Specify in config file, e.g., ATAX.config>>>"
    calculation_method = \
      "Das Ergebnis wurde gemäß Par.4/3 EStG nach der Nettomethode erstellt"

    def __init__ (self, name = "", year = 0, konto_desc = None, vst_korrektur = 1.0) :
        Account.__init__ (self, name, vst_korrektur)
        self.year                           = year or \
             day_to_time_tuple ("1.1").year - 1
        self.konto_desc                     = konto_desc or {}
        self.soll_saldo                     = defaultdict (EUR)
        self.haben_saldo                    = defaultdict (EUR)
        self.ausgaben                       = defaultdict (EUR)
        self.einnahmen                      = defaultdict (EUR)
        self.vorsteuer                      = defaultdict (EUR)
        self.vorsteuer_EUst                 = defaultdict (EUR)
        self.ust                            = defaultdict (EUR)
        self.ust_igE                        = defaultdict (EUR)
        self.ust_revCharge                  = defaultdict (EUR)
        self.buchung_zahl                   = defaultdict (int)
        self.ausgaben_total                 = EUR (0)
        self.einnahmen_total                = EUR (0)
        self.vorsteuer_total                = EUR (0)
        self.ust_total                      = EUR (0)
        self.k_entries                      = defaultdict (list)
        self.kblatt                         = defaultdict (list)
    # end def __init__

    def add (self, entry) :
        """Add `entry' to account `self'."""
        assert (isinstance (entry, Account_Entry))
        self.buchung_zahl [entry.konto]       += 1
        self.buchung_zahl [entry.gegen_konto] += 1
        self.kblatt       [entry.konto].append (entry.kontenzeile ())
        self.k_entries    [entry.konto].append (entry)
        if "-" in entry.dir :
            self.soll_saldo  [entry.konto]       += entry.soll_betrag
            self.haben_saldo [entry.gegen_konto] += entry.soll_betrag
            if  (  (not self.t_konto_ignore_pat.match (entry.konto))
                or ("u" in entry.cat)
                ) :
                betrag = self._effective_amount (entry, entry.soll_betrag)
                self._add_ausgabe  (entry, betrag, entry.vat)
            else :
                pass #print "%s not added to self.ausgaben" % entry.konto
        elif "+" in entry.dir :
            self.haben_saldo [entry.konto]       += entry.haben_betrag
            self.soll_saldo  [entry.gegen_konto] += entry.haben_betrag
            if  (  (not self.t_konto_ignore_pat.match (entry.konto))
                or ("u" in entry.cat)
                ) :
                betrag = self._effective_amount (entry, entry.haben_betrag)
                self._add_einnahme (entry, betrag, entry.vat)
            else :
                pass #print "%s not added to self.einnahmen" % entry.konto
    # end def add

    def _effective_amount (self, entry, amount) :
        if  (    self.t_konto_ignore_pat.match (entry.konto)
            and "u" in entry.cat
            ) or entry.konto in self.ignore :
            return 0
        else :
            return amount
    # end def _effective_amount

    def _add_ausgabe (self, entry, betrag, vat) :
        if "u" in entry.cat :
            if "i" in entry.cat :
                vat_dict = self.ust_igE
            elif "r" in entry.cat :
                vat_dict = self.ust_revCharge
            elif "z" in entry.cat :
                vat_dict = self.vorsteuer_EUst
            else :
                vat_dict = self.vorsteuer
            vat_dict      [entry.month] += vat
            self.vorsteuer_total        += vat
        self.ausgaben [entry.konto] += betrag
        self.ausgaben_total         += betrag
    # end def _add_ausgabe

    def _add_einnahme (self, entry, betrag, vat) :
        if "u" in entry.cat :
            if "i" in entry.cat :
                ust_dict = self.ust_igE
            elif "r" in entry.cat :
                ust_dict = self.ust_revCharge
            else :
                if "z" in entry.cat :
                    pyk.fprint \
                        ("*** Einnahme cannot include Einfuhrumsatzsteuer")
                    pyk.fprint (entry.line)
                ust_dict = self.ust
            ust_dict       [entry.month] += vat
            self.ust_total               += vat
        self.einnahmen [entry.konto] += betrag
        self.einnahmen_total         += betrag
     # end def _add_einnahme

    def finish (self) :
        if not self._finished :
            self._finished = True
            self._do_gkonto \
                ( self.vorsteuer,   self.vorsteuer_gkonto
                , self.soll_saldo,  "Vorsteuer"
                , lambda s : (s, 0)
                )
            self._do_gkonto \
                ( self.vorsteuer_EUst, self.eust_gkonto
                , self.soll_saldo,     "Einfuhrumsatzsteuer"
                , lambda s : (s, 0)
                )
            self._do_gkonto \
                ( self.ust,            self.ust_gkonto
                , self.haben_saldo,    "Umsatzsteuer"
                , lambda s : (0, s)
                )
            self._do_gkonto \
                ( self.ust_igE,        self.ige_gkonto
                , self.soll_saldo,     "Vor- und Umsatzsteuer igE"
                , lambda s : (s, s)
                , self.haben_saldo
                )
            self._do_gkonto \
                ( self.ust_revCharge,  self.rvc_gkonto
                , self.soll_saldo,     "Vor- und Umsatzsteuer rev. Ch."
                , lambda s : (s, s)
                , self.haben_saldo
                )
            for k in sorted (self.kblatt) :
                if k in self.privat :
                    pa      = self.privat [k]
                    factor  = pa / 100.0
                    p_soll  = self.soll_saldo  [k] * factor
                    p_haben = self.haben_saldo [k] * factor
                    p_entry = self._fix_privat_anteil \
                        (k, pa, factor, p_soll, p_haben)
                    k_desc  = self.konto_desc.get (k, "")
                    vat_pd  = self.vat_privat_default
                    self.ausgaben   [k] *= (1 - factor)
                    self.einnahmen  [k] *= (1 - factor)
                    self.konto_desc [k]  = "%s (abz. %s)" % \
                        (k_desc, p_entry.desc)
                    p_vat = p_soll * (self.vat_privat.get (k, vat_pd) - 1.00)
                    self._fix_ust_privat (k, k_desc, p_vat, p_entry.desc)
                else :
                    for ke in self.k_entries [k] :
                        if ke.p_konto in self.privat :
                            pa      = self.privat [ke.p_konto]
                            factor  = pa / 100.0
                            p_soll  = ke.soll_betrag  * factor
                            p_haben = ke.haben_betrag * factor
                            self.ausgaben  [k] -= p_soll
                            self.einnahmen [k] -= p_haben
                            self._fix_privat_anteil \
                                (k, pa, factor, p_soll, p_haben, ke.desc)
    # end def finish

    def _do_gkonto (self, ust, gkonto, saldo, txt, soll_haben, saldo2 = None) :
        for m in sorted (ust) :
            if ust.get (m, 0) != 0 :
                self.buchung_zahl [gkonto] += 1
                saldo             [gkonto] += ust [m]
                if saldo2 is not None :
                    saldo2        [gkonto] += ust [m]
                s, h                        = soll_haben (ust [m])
                self.kblatt [gkonto].append \
                    (Ust_Gegenbuchung (m, gkonto, s, h, txt).kontenzeile ())
    # end def _do_gkonto

    def _fix_privat_anteil (self, k, pa, factor, p_soll, p_haben, desc = "") :
        p_desc  = ("%2d%% Privatanteil %s" % (int (pa + 0.5), desc)).strip ()
        p_entry = Privatanteil ("9200", -p_soll, -p_haben, p_desc)
        self.soll_saldo  [k] -= p_soll
        self.haben_saldo [k] -= p_haben
        self.kblatt [k].append (p_entry.kontenzeile ())
        return p_entry
    # end def _fix_privat_anteil

    def _fix_ust_privat (self, k, k_desc, p_vat, p_desc) :
        p_desc = "%s [%s (%s)]" % (p_desc, k, k_desc)
        gkonto = self.ust_gkonto
        self.buchung_zahl [gkonto] += 1
        self.haben_saldo  [gkonto] += p_vat
        self.kblatt [gkonto].append \
            (Privatanteil ("9200", 0, p_vat, p_desc, gkonto).kontenzeile ())
    # end def _fix_ust_privat

    def print_konten (self) :
        self.finish ()
        tc = EUR.target_currency.name
        for k in sorted (self.kblatt) :
            head  = "%s    %s" % (self.year, self.konto_desc.get (k, "")) [:64]
            tail  = "Konto-Nr. %5s" % k
            belly = " " * (79 - len (head) - len (tail))
            pyk.fprint ("\n\n%s%s%s" % (head, belly, tail))
            self.print_sep_line ()
            pyk.fprint \
                ( "%-4s %-6s  %-35s%-3s  %12s  %12s"
                % ("Tag", "Gegkto", "Text", "Ust", "Soll", "Haben")
                )
            self.print_sep_line ()
            pyk.fprint ("\n".join (self.kblatt [k]))
            pyk.fprint \
                ( "\n%12s %-38s  %12s  %12s"
                % ( "", "Kontostand neu"
                  , self.soll_saldo  [k].as_string_s ()
                  , self.haben_saldo [k].as_string_s ()
                  )
                )
            pyk.fprint \
                ( "\n%12s %-48s  %12s %s"
                % ( ""
                  , "Saldo      neu"
                  , (self.soll_saldo [k] - self.haben_saldo [k]).as_string_s ()
                  , tc
                  )
                )
            self.print_sep_line ()
    # end def print_konten

    def print_konto_summary (self) :
        self.finish ()
        tc = EUR.target_currency.name
        pyk.fprint ("Zusammenfassung Konten %-52s %s" % (self.year, tc))
        self.print_sep_line ()
        pyk.fprint \
            ( "%-5s %12s %12s  %12s %s"
            % ("Konto", "Soll", "Haben", "Saldo", "Text")
            )
        self.print_sep_line ()
        for k in sorted (self.kblatt) :
            pyk.fprint \
                ( "%-5s%13s%13s %13s %s"
                % ( k
                  , self.soll_saldo  [k].as_string_s ()
                  , self.haben_saldo [k].as_string_s ()
                  , (self.soll_saldo [k] - self.haben_saldo [k]).as_string_s ()
                  , self.konto_desc.get (k, "") [:33]
                  )
                )
    # end def print_konto_summary

    def print_sep_line (self) :
        pyk.fprint ("%s" % ("=" * 79, ))
    # end def print_sep_line

    def print_ein_aus_rechnung (self) :
        self.finish ()
        tc = EUR.target_currency.name
        pyk.fprint (self.firma)
        pyk.fprint \
            ( underlined
                ("Einnahmen/Ausgabenrechnung %s (%s)" % (self.year, tc))
            )
        pyk.fprint ("\n")
        pyk.fprint (underlined ("Betriebseinnahmen"))
        pyk.fprint ("\n")
        format  = "%-40s %15s %15s"
        format1 = "%-40s -%14s %15s"
        e_total = EUR (0)
        for k in sorted (self.einnahmen) :
            einnahmen = self.einnahmen [k] - self.ausgaben [k]
            if k [0] not in ("4", "8") : continue
            if einnahmen == 0          : continue
            e_total = e_total + einnahmen
            pyk.fprint \
                ( format
                % ( self.konto_desc.get (k, "") [:40], ""
                  , einnahmen.as_string_s ()
                  )
                )
        pyk.fprint (format % ("", "", "_" * 15))
        pyk.fprint (format % ("", "", e_total.as_string_s ()), tc)
        pyk.fprint ("\n")
        pyk.fprint (underlined ("Betriebsausgaben"))
        pyk.fprint ("\n")
        a_total = EUR (0)
        for k in sorted (self.ausgaben) :
            ausgaben = self.ausgaben [k] - self.einnahmen [k]
            if k [0] not in ("5", "6", "7") : continue
            if ausgaben == 0                : continue
            a_total = a_total + ausgaben
            pyk.fprint \
                ( format
                % ( self.konto_desc.get (k, "") [:40]
                  , ausgaben.as_string_s (), ""
                  )
                )
        if self.vst_korrektur != 1 :
            p_anteil = a_total * (1 - self.vst_korrektur)
            pyk.fprint (format  % ( "", "_" * 15, ""))
            pyk.fprint (format  % ( "", a_total.as_string_s (), ""))
            pyk.fprint \
                ( format1
                % ( "Privatanteil  %5.2f%%" % ((1 - self.vst_korrektur) * 100, )
                  , p_anteil.as_string_s (), ""
                  )
                )
            if self.gewerbe_anteil :
                self.g_anteil = g_anteil = a_total * self.gewerbe_anteil
                pyk.fprint \
                    ( format1
                    % ( "Gewerbeanteil %5.2f%%" % (self.gewerbe_anteil * 100, )
                      , g_anteil.as_string_s (), ""
                      )
                    )
            else :
                g_anteil = 0
            a_total = a_total - p_anteil - g_anteil
        pyk.fprint (format  % ( "", "_" * 15, "_" * 15))
        pyk.fprint \
            ( format1
            % ( ""
              , a_total.as_string_s ()
              , (e_total - a_total).as_string_s ()
              )
            , tc
            )
        pyk.fprint ("\n%s." % (self.calculation_method, ))
    # end def print_ein_aus_rechnung

    g_anteil = EUR (0)

# end class T_Account

class H_Account_Entry (Account_Entry) :
    """Entry of H_Account"""

    Ancestor = __Ancestor = Account_Entry

    def __init__ (self, * args, ** kw) :
        self.__Ancestor.__init__ (self, * args, ** kw)
        if "-" in self.dir and self.vat_p != 1 :
            self.netto       += self.vat_vstk
            self.soll_betrag += self.vat_vstk
    # end def __init__

    def kontenzeile (self) :
        try :
            return \
                ( "\\Einzelposten{%s\\hfill %s}{%s%s}"
                % (self.desc, self.date, self.dir, self.netto.as_string ())
                )
        except Exception as exc :
            pyk.fprint (exc)
            pyk.fprint (sorted (self.__dict__.items ()))
            raise
    # end def kontenzeile

# end class H_Account_Entry

class H_Account (T_Account) :
    """House account for a specific time interval."""

    Ancestor = __Ancestor = T_Account
    Entry    = H_Account_Entry
    ignore             = () ### DON'T IGNORE ANYTHING HERE
    t_konto_ignore_pat = Regexp (r"^DON'T MATCH ANYTHING HERE$", re.X)

    def _effective_amount (self, entry, amount) :
        return amount
    # end def _effective_amount

# end class H_Account

class Konto_Desc (dict) :
    """Model kontenplan of austrian accounting system"""

    def __init__ (self, file) :
        dict.__init__ (self)
        k_d = self
        d_k = self.reverse = {}
        if isinstance (file, str) :
            file = open (file, "rb")
        pat   = Regexp (r"^[0-9]")
        s_pat = Regexp (r"\s*:\s*")
        for l in file :
            line = pyk.decoded (l, "utf-8", "iso-8859-15")
            if not pat.match (line) :
                continue
            (konto, desc)  = s_pat.split   (line)
            konto          = konto.replace ("_", "0").strip ()
            desc           = desc.replace  ('"', "" ).strip ()
            k_d [konto]    = desc
            d_k [desc]     = konto
    # end def __init__

# end class Konto_Desc

class _ATAX_Command_ (TFL.Meta.Object) :
    """Main class for accounting scripts"""

    _real_name          = "Command"
    default_categories  = "u"
    max_args            = -1
    min_args            = 0

    def __init__ (self) :
        cmd = self._command ()
        self.load_config (cmd)
        if cmd.all :
            categories  = "."
        else :
            categories  = "[" + "".join (cmd.categories) + "]"
        categories      = Regexp (categories)
        source_currency = cmd.source_currency
        vst_korrektur   = cmd.vst_korrektur
        account         = self._create_account \
            (cmd, categories, source_currency, vst_korrektur)
        self._add_files (cmd, account, categories, source_currency)
        self._output    (cmd, account, categories, source_currency)
    # end def __init__

    def _command (self) :
        cmd = TFL.CAO.Cmd \
            ( args        = self._arg_spec ()
            , opts        = self._opt_spec ()
            , min_args    = self.min_args
            , max_args    = self.max_args
            , description = self.__doc__ or ""
            )
        return cmd ()
    # end def _command

    @classmethod
    def load_config (cls, cmd) :
        globs   = globals ()
        cf_dict = dict    (ATAX = ATAX)
        for cf in cmd.Config :
            TFL.load_config_file (cf, globs, cf_dict)
    # end def load_config

    def _add_files (self, cmd, account, categories, source_currency) :
        if cmd.stdin :
            assert cmd.argn == 0
            account.add_lines    (sys.stdin, categories, source_currency)
        else :
            for f_or_p in cmd.argv :
                for file_name in glob.glob (f_or_p) :
                    account.add_file (file_name, categories, source_currency)
    # end def _add_files

    def _arg_spec (self) :
        return ("file:P", )
    # end def _arg_spec

    def _opt_spec (self) :
        return \
            ( "-all:B"
            , "-categories:S,=%s" % self.default_categories
            , "-Config:P,?Config file(s)"
            , "-stdin:B?Read data from standard input"
            , "-vst_korrektur:F=1.0"
            , TFL.CAO.Arg.EUC_Source      ()
            , TFL.CAO.Arg.EUC_Target      ()
            , TFL.CAO.Opt.Output_Encoding (default = "utf-8")
            )
    # end def _opt_spec

Command = _ATAX_Command_ # end class

if __name__ != "__main__" :
    ATAX._Export ("*")
### __END__ ATAX.accounting



# -*- coding: utf-8 -*-
# Copyright (C) 2008-2015 Mag. Christian Tanzer. All rights reserved
# Glasauergasse 32, A--1130 Wien, Austria. tanzer@swing.co.at
# ****************************************************************************
# This module is part of the package ATAX.
#
# This module is licensed under the terms of the BSD 3-Clause License
# <http://www.c-tanzer.at/license/bsd_3c.html>.
# ****************************************************************************
#
#++
# Name
#    ATAX.fahrtenbuch
#
# Purpose
#    Fahrtenbuch
#
# Revision Dates
#     4-Jan-2008 (CT) Creation (ported from a perl version)
#     5-Jan-2008 (CT) Use `CAL.Date_Time` instead of (obsolete) `TFL.Date_Time`
#     9-Jul-2009 (MG) Support for `km_geld` added
#     5-Aug-2009 (CT) `km_business`, `km_private`, and `private_percent`
#                     factored (in class `Fahrtenbuch`)
#     3-Jan-2010 (CT) Use `TFL.CAO` instead of `TFL.Command_Line`
#    29-Oct-2015 (CT) Improve Python 3 compatibility
#    31-Oct-2015 (CT) Add `@pyk.adapt__str__` to `FB_Entry`
#    28-Dec-2015 (CT) Fix `__str__` (use `pyk.text_type`, not `str`)
#    ««revision-date»»···
#--

from   __future__  import absolute_import
from   __future__  import division
from   __future__  import print_function
from   __future__  import unicode_literals

from   _ATAX             import ATAX
from   _CAL              import CAL
from   _TFL              import TFL

from   _ATAX.accounting  import ignor_pat

from   _TFL.predicate    import *
from   _TFL.pyk          import pyk
from   _TFL.Regexp       import *
from   _TFL._Meta.Once_Property import Once_Property

import _CAL.Date_Time

import _TFL._Meta.Object
import _TFL.CAO
import _TFL.Environment

@pyk.adapt__bool__
@pyk.adapt__str__
class FB_Entry (TFL.Meta.Object) :
    """Model one entry of a Fahrtenbuch"""

    str_format  = r"  %-11s & %8d & %6d & %s"
    tex_format  = r"  %-10s & %6d & %6d & %4d & %4d & \sf %s"
    atax_format = \
        " %s  & & & %6.2f & b & 7500 & 2100 & -  & e & & KM Geld (%5.1f km)"

    def __init__ (self, date, km_start, km_finis, priv, desc, ** kw) :
        self.date     = date
        self.km_start = float (km_start)
        self.km_finis = float (km_finis)
        self.priv     = float (priv or 0)
        self.desc, _, self.comment = (x.strip () for x in split_hst (desc, "%"))
        self.__dict__.update (kw)
    # end def __init__

    @Once_Property
    def delta (self) :
        return self.km_finis - self.km_start
    # end def delta

    @Once_Property
    def km_business (self) :
        return self.delta * (1.0 - self.priv / 100.)
    # end def km_business

    @Once_Property
    def km_private (self) :
        return self.delta * (self.priv / 100.)
    # end def km_private

    def tex (self) :
        date    = self.date.formatted ("%d.%m.%Y")
        bus_km  = self.km_business
        priv_km = self.km_private
        desc    = self.desc ### TeX-quote `desc`
        return self.tex_format % \
            (date, self.km_start, self.km_finis, bus_km, priv_km, desc)
    # end def tex

    def atax (self) :
        date    = self.date.formatted ("%d.%m")
        km      = self.km_business
        f       = 0.42 ### 0.42 Euro/km
        if self.date < CAL.Date_Time (2008, 7, 1) :
            f   = 0.38
        return self.atax_format % (date, f * km, km)
    # end def atax

    def __bool__ (self) :
        return bool (self.delta)
    # end def __bool__

    def __str__ (self) :
        date = self.date.formatted ("%d.%m.%Y")
        return self.str_format % (date, self.km_finis, self.priv, self.desc)
    # end def __str__

# end class FB_Entry

class Fahrtenbuch (TFL.Meta.Object) :
    """Model a Fahrtenbuch"""

    lines_per_page = 50
    Entry          = FB_Entry

    def __init__ (self, user, entries = []) :
        self.user    = user
        self.entries = []
        for e in entries :
            self.add (e)
    # end def __init__

    def add (self, entry) :
        self.entries.append (entry)
    # end def add

    @classmethod
    def from_file (cls, user, file_name) :
        result = cls (user)
        add    = result.add
        last   = None
        with open (file_name, "rb") as file :
            for l in result._read_lines (file) :
                line = pyk.decoded (l, "utf-8", "iso-8859-1")
                try :
                    d, km, priv, desc = \
                        [f.strip () for f in line.split ("&", 4)]
                except ValueError :
                    pyk.fprint ("Split error `%s`" % line)
                else :
                    last = result._new_entry (last, d, km, priv, desc)
                    add (last)
        return result
    # end def from_file

    @property
    def km_business (self) :
        return sum (e.km_business for e in self.entries)
    # end def km_business

    def km_geld (self) :
        result = ["""$source_currency = "EUR";"""]
        for e in self.entries :
            if e.km_business :
                result.append (e.atax ())
        return "\n".join (result)
    # end def km_geld

    @property
    def km_private (self) :
        return sum (e.km_private for e in self.entries)
    # end def km_private

    @property
    def private_percent (self) :
        kmb    = self.km_business
        kmp    = self.km_private
        result = 0
        if kmp :
            result = 100.0 / ((kmb + kmp) / kmp)
        return result
    # end def private_percent

    def tex (self) :
        result  = []
        entries = self.entries
        if entries :
            head    = entries [0]
            tail    = entries [-1]
            h_date  = head.date.formatted ("%m/%Y")
            t_date  = tail.date.formatted ("%m/%Y")
            dates   = h_date
            if h_date != t_date :
                dates = "%s -- %s" % (h_date, t_date)
            add    = result.append
            add ( ( r"\def\fahrtenbuchpageheader"
                    r"{\begin{Larger}\bf "
                    r"\strut\triline{Fahrtenbuch}{%s}{%s}"
                    r"\end{Larger}"
                    r"\par\vspace{0.0cm}"
                    r"}"
                  )
                % (self.user, dates)
                )
            add (r"\begin{fahrtenbuch}")
            i   = 1
            lpp = self.lines_per_page
            for e in entries :
                if e :
                    sep = [r"\\ ", r"\CR"] [(i % 5) == 0]
                    add ("%s %s" % (e.tex (), sep.strip ()))
                    i += 1
                    if i > lpp :
                        add (r"\end{fahrtenbuch}")
                        add (r"\\ \hbox{\strut}\hfill ./..%")
                        add (r"\begin{fahrtenbuch}")
                        i = 1
            add (r"\hline\hline")
            kmb  = self.km_business
            kmp  = self.km_private
            priv = self.private_percent
            add ( FB_Entry.tex_format
                % ( "Total", head.km_start, tail.km_finis, kmb, kmp
                  , r"\hfill Privatanteil = \percent{%5.2f} \CR" % priv
                  )
                )
            add (r"\end{fahrtenbuch}")
        return "\n".join (result)
    # end def tex

    def _new_entry (self, last, d, km_finis, priv, desc) :
        km_start = km_finis
        if last is not None :
            km_start = last.km_finis
        return self.Entry \
            (CAL.Date_Time.from_string (d), km_start, km_finis, priv, desc)
    # end def _new_entry

    def _read_lines (self, file) :
        for l in file :
            line = pyk.decoded (l.strip (), "utf-8", "iso-8859-1")
            if line and not ignor_pat.match (line) :
                yield line
    # end def _read_lines

    def __str__ (self) :
        return "\n".join (pyk.text_type (e) for e in self.entries)
    # end def __str__

# end class Fahrtenbuch

def _main (cmd) :
    ATAX.Command.load_config (cmd)
    fb = Fahrtenbuch.from_file (cmd.user, cmd.fahrtenbuch)
    if not cmd.km_geld :
        pyk.fprint (fb.tex ())
    else :
        pyk.fprint (fb.km_geld ())
# end def _main

_Command = TFL.CAO.Cmd \
    ( handler     = _main
    , args        =
        ("fahrtenbuch:P?File defining fahrtenbuch", )
    , min_args    = 1
    , max_args    = 1
    , opts        =
        ( "Config:P,?Config file(s)"
        , "user:S=%s" % TFL.Environment.username.capitalize ()
        , "km_geld:B?Print the factoring for the KM Geld"
        , TFL.CAO.Opt.Output_Encoding (default = "iso-8859-15")
        )
    )

if __name__ != "__main__" :
    ATAX._Export ("*")
else :
    _Command ()
### __END__ ATAX.fahrtenbuch



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2016 Mag. Christian Tanzer. All rights reserved
# Glasauergasse 32, A--1130 Wien, Austria.
# Web: http://www.c-tanzer.at/en/ Email: tanzer@swing.co.at
# All rights reserved
# ****************************************************************************
# This module is licensed under the terms of the BSD 3-Clause License
# <http://www.c-tanzer.at/license/bsd_3c.html>.
# ****************************************************************************
#
#++
# Name
#    _ATAX.setup
#
# Purpose
#    Setup file for package namespace ATAX
#
# Revision Dates
#    11-Oct-2016 (CT) Creation
#    ««revision-date»»···
#--

from   __future__               import print_function

from   codecs                   import open
from   setuptools               import setup, Command

import ast
import os
import re
import subprocess               as     SUBP
import sys

_version_re = re.compile(r'__version__\s+=\s+(.*)')

src_dir = os.path.dirname (__file__)
if src_dir :
    os.chdir (src_dir)

license = "BSD License"

name    = "ATAX"
p_name  = "_ATAX"

with open ("__init__.py", encoding = "utf-8") as f :
    version = str \
        (ast.literal_eval (_version_re.search (f.read ()).group (1)))

packages = []
for root, dirs, files in os.walk (os.path.join ('..', p_name)) :
    # ugly: need to modify dirs in place!
    dd = set (d for d in dirs if d.startswith ('_') and d != "__pycache__")
    for d in dirs :
        if d not in dd :
            dirs.remove (d)
    packages.append (root.replace ('/', '.').strip ("."))

data_files = ["LICENSE", "README.rst"]
data_dirs  = []
with SUBP.Popen \
    (["find", ".", "-name", "-I18N"], stdout=SUBP.PIPE).stdout as pipe :
    found = pipe.read ().strip ()
    if found :
        i18n_dirs = found.split ("\n")
        data_dirs.extend (i18n_dirs)
for d in data_dirs :
    with SUBP.Popen \
        (["find", d, "-type", "f"], stdout=SUBP.PIPE).stdout as pipe :
        found = pipe.read ().strip ()
        if found :
            files = found.split ("\n")
            data_files.extend (files)

class Test_Command (Command) :
    user_options = []

    def initialize_options (self) :
        pass

    def finalize_options (self) :
        pass

    def run (self) :
        import _TFL.run_doctest
        _TFL.run_doctest.Command (["-summary", "-transitive", "./"])

# end class Test_Command

setup \
    ( name                 = name
    , version              = version
    , description          =
        "Package for doing Einnahmen-Ausgabenrechnung according to "
        "Austrian tax laws."
    , long_description     =
        open ("README.rst", encoding = "utf-8").read ().strip ()
    , license              = license
    , author               = "Christian Tanzer"
    , author_email         = "tanzer@swing.co.at"
    , url                  = "https://github.com/Tapyr/tapyr"
    , packages             = packages
    , package_dir          = { p_name : "../" + p_name }
    , package_data         = { p_name : data_files     }
    , platforms            = 'Any'
    , classifiers          = \
        [ 'Development Status :: 5 - Production/Stable'
        , 'License :: OSI Approved :: ' + license
        , 'Operating System :: OS Independent'
        , 'Programming Language :: Python'
        , 'Programming Language :: Python :: 2'
        , 'Programming Language :: Python :: 2.7'
        , 'Programming Language :: Python :: 3'
        , 'Programming Language :: Python :: 3.5'
        , 'Intended Audience :: Developers'
        , 'Topic :: Software Development :: Libraries :: Python Modules'
        ]
    , install_requires     = ["TFL", "CAL"]
    , extras_require           = dict ()
    , cmdclass             = dict (test = Test_Command)
    , zip_safe             = False ### no eggs, please
    )



# -*- coding: utf-8 -*-
# Copyright (C) 1999-2015 Mag. Christian Tanzer. All rights reserved
# Glasauergasse 32, A--1130 Wien, Austria. tanzer@swing.co.at
# ****************************************************************************
# This module is part of the package ATAX.
#
# This module is licensed under the terms of the BSD 3-Clause License
# <http://www.c-tanzer.at/license/bsd_3c.html>.
# ****************************************************************************
#
#++
# Name
#    umsatzsteuer
#
# Purpose
#    Calculate monthly VAT for the austrian finance ministry
#
# Revision Dates
#    14-Jan-1999 (CT) Creation
#     7-Feb-1999 (CT) Special handling for `source_currency' added to
#                     `code_pat' processing
#    26-Feb-1999 (CT) `help_on_err = 1' passed to Command_Line
#     4-Mar-1999 (CT) Handling of `privat' added
#     3-Jan-2000 (CT) Logic of `umsatzsteuer' factored into
#                     `V_Account.add_lines'
#    26-Dec-2001 (CT) Access to cmd-options simplified
#    26-Dec-2001 (CT) Function `umsatzsteuer` moved to `accounting`
#     6-Jan-2003 (CT) Alias `umsatzsteuer` for `add_account_file` added
#    19-Feb-2006 (CT) Import from package _ATAX
#    15-May-2006 (CT) Call to `finish` added
#    11-Feb-2007 (CT) `string` functions replaced by `str` methods
#    17-Sep-2007 (CT) Modernized
#    17-Sep-2007 (CT) Use `Account.add_file`
#    17-Sep-2007 (CT) `main` refactored
#     3-Jan-2010 (CT) Use `TFL.CAO` instead of `TFL.Command_Line`
#    29-Oct-2015 (CT) Improve Python 3 compatibility
#    ««revision-date»»···
#--

from   __future__  import absolute_import
from   __future__  import division
from   __future__  import print_function
from   __future__  import unicode_literals

from _ATAX.accounting import *

sep_1000 = { "." : ",", "," : "."}

class Ust_Command (Command) :

    def _add_files (self, cmd, account, categories, source_currency) :
        self.__super._add_files (cmd, account, categories, source_currency)
        account.finish ()
    # end def _add_files

    def _create_account (self, cmd, categories, source_currency, vst_korrektur) :
        if cmd.online_format :
            EUC.target_currency.decimal_sign = ","
            EUC.target_currency.sep_1000     = ""
        else :
            EUC.target_currency.decimal_sign = cmd.decimal_sign
            EUC.target_currency.sep_1000     = sep_1000 [cmd.decimal_sign]
        return V_Account (vst_korrektur = vst_korrektur)
    # end def _create_account

    def _output (self, cmd, account, categories, source_currency) :
        if cmd.online_format :
            account.print_summary_online ()
        else :
            if not cmd.summary :
                account.print_entries ("\f")
                if cmd.reverse :
                    account.print_entries_by_group ("\f")
            account.print_summary ()
    # end def _output

    def _opt_spec (self) :
        return self.__super._opt_spec () + \
            ( "-decimal_sign:S=."
            , "-online_format:B?Format for cope/paste into online form"
            , "-reverse:B"
            , "-summary:B"
            )
    # end def _opt_spec

Command = Ust_Command # end class Ust_Command

if __name__ == "__main__":
    Ust_Command ()
### __END__ ATAX.umsatzsteuer
